# ------------------------------------------------------------------------------
# Name:          meimetadata.py
# Purpose:       MeiMetadata is an object that takes a music21 Metadata
#                object, and computes the MEI structure for it.
#
# Authors:       Greg Chapman <gregc@mac.com>
#
# Copyright:     (c) 2023 Greg Chapman
# License:       MIT, see LICENSE
# ------------------------------------------------------------------------------
import sys
from xml.etree.ElementTree import TreeBuilder  # , Element
import typing as t

import music21 as m21

# from converter21.mei import MeiExportError
# from converter21.mei import MeiInternalError
from converter21.mei import MeiElement
from converter21.shared import M21Utilities
from converter21.shared import SharedConstants

environLocal = m21.environment.Environment('converter21.mei.meimetadata')

# For debug or unit test print, a simple way to get a string which is the current function name
# with a colon appended.
# for current func name, specify 0 or no argument.
# for name of caller of current func, specify 1.
# for name of caller of caller of current func, specify 2. etc.
# pylint: disable=protected-access
funcName = lambda n=0: sys._getframe(n + 1).f_code.co_name + ':'  # pragma no cover
# pylint: enable=protected-access

class MeiMetadataItem:
    def __init__(self, m21Item: tuple[str, t.Any], md: m21.metadata.Metadata) -> None:
        self.key: str = m21Item[0]
        self.value: t.Any = m21Item[1]

        # self.uniqueName is only set if self.key is a standard music21 metadata
        #   unique name.
        # self.humdrumRefKey is set in two cases:
        #   1. self.key is 'humdrum:XXX' and 'XXX' is something we understand
        #       (we share this knowledge with our humdrum converter).
        #   2. self.key is a standard music21 uniqueName, and it maps to a
        #       particular humdrum key (this is knowledge shared with our
        #       humdrum converter).

        self.uniqueName: str = ''
        self.humdrumRefKey: str = ''
        self.meiMetadataKey: str = ''
        if md._isStandardUniqueName(self.key):
            self.uniqueName = self.key
            self.humdrumRefKey = (
                M21Utilities.m21MetadataPropertyUniqueNameToHumdrumReferenceKey.get(self.key, '')
            )
            if self.uniqueName == 'otherContributor' and not self.humdrumRefKey:
                if t.TYPE_CHECKING:
                    assert isinstance(self.value, m21.metadata.Contributor)
                role: str = self.value.role
                self.humdrumRefKey = (
                    M21Utilities.m21OtherContributorRoleToHumdrumReferenceKey.get(role, '')
                )

        elif self.key.startswith('humdrum:'):
            hdKey: str = self.key[8:]
            if hdKey in M21Utilities.validHumdrumReferenceKeys:
                self.humdrumRefKey = hdKey
        elif self.key.startswith('mei:'):
            if self.key in M21Utilities.validMeiMetadataKeys:
                self.meiMetadataKey = self.key

        # self.isCustom is True if we don't have a uniqueName or humdrumRefKey or meiMetadataKey
        self.isCustom: bool = (
            not self.uniqueName and not self.humdrumRefKey and not self.meiMetadataKey
        )

        self.isContributor: bool = md._isContributorUniqueName(self.uniqueName)

        # convert self.value to an appropriate string (no html escaping yet)
        self.meiValue: str = M21Utilities.m21MetadataValueToString(self.value, lineFeedOK=True)

        # Add any other attributes set by converter21 importers (e.g. MEI importer
        # will have set value.meiVersion to 'version', which we will translate to
        # @version="version")
        self.meiOtherAttribs: dict[str, str] = {}
        otherAttribs: set[str] = getattr(self.value, 'c21OtherAttribs', set())
        for attrib in otherAttribs:
            if attrib in ('meiVersion', 'humdrumVersion'):
                version: str = getattr(self.value, attrib, '')
            if version:
                self.meiOtherAttribs['version'] = version

class MeiMetadata:
    def __init__(self, m21Metadata: m21.metadata.Metadata) -> None:
        self.m21Metadata: m21.metadata.Metadata = m21Metadata
        # self.contents is structured just like m21Metadata._contents v8 and later,
        # except that the key is the humdrum ref key (if there is one), or the
        # uniqueName (if there is one), or the custom key if that's all there is.
        self.contents: dict[str, list[MeiMetadataItem]] = {}
        for m21Item in m21Metadata.all(returnPrimitives=True, returnSorted=False):
            meiItem = MeiMetadataItem(m21Item, m21Metadata)
            key: str = (
                meiItem.humdrumRefKey
                or meiItem.uniqueName
                or meiItem.meiMetadataKey
                or meiItem.key
            )
            currList: list[MeiMetadataItem] | None = self.contents.get(key, None)
            if currList is None:
                self.contents[key] = [meiItem]
            else:
                currList.append(meiItem)

        # The xml:id of the main work (the work that is actually encoded).
        # This will be referenced in mdiv@decls (or maybe music@decls).
        self.mainWorkXmlId: str = ''

        # Saved off state during makeRootElement (e.g. MeiElement trees that
        # may need to be re-used).

        # These are the simplified title and composers, that are used in
        # <fileDesc> and <source>.  The main <work> will have ALL the details.
        self.simpleTitleElement: MeiElement | None = None
        self.simpleComposerElements: list[MeiElement] = []

        # This element contains any <mads> elements that will end up in <work>/<extMeta>.
        self.madsCollection: MeiElement | None = None

    def makeRootElement(self, tb: TreeBuilder):
        meiHead: MeiElement = MeiElement('meiHead')
        fileDesc: MeiElement = self.makeFileDescElement()
        encodingDesc: MeiElement | None = self.makeEncodingDescElement()
        workList: MeiElement | None = self.makeWorkListElement()
        manifestationList: MeiElement | None = self.makeManifestationListElement()
        extMeta: MeiElement | None = self.makeExtMetaElementForMeiHead()

        # fileDesc is required
        meiHead.subElements.append(fileDesc)

        if encodingDesc:
            meiHead.subElements.append(encodingDesc)
        if workList:
            meiHead.subElements.append(workList)
        if manifestationList:
            meiHead.subElements.append(manifestationList)
        if extMeta:
            meiHead.subElements.append(extMeta)

        meiHead.makeRootElement(tb)

    def anyExist(self, *args: str) -> bool:
        for arg in args:
            if self.contents.get(arg, []):
                return True
        return False

    def makeFileDescElement(self) -> MeiElement:
        # meiHead/fileDesc is required, so we never return None here
        fileDesc: MeiElement = MeiElement('fileDesc')

        # titleStmt is required, so we create/append in one step
        titleStmt: MeiElement = fileDesc.appendSubElement('titleStmt')

        if self.simpleTitleElement is None:
            self.simpleTitleElement = self.makeSimpleTitleElement()
        if self.simpleTitleElement is not None:
            titleStmt.subElements.append(self.simpleTitleElement)

        # pubStmt describes the MEI file we are writing.  It is, of course,
        # unpublished, so we use <unpub> to say that.  This pubStmt will
        # always be non-empty, so create/append in one step
        pubStmt: MeiElement = fileDesc.appendSubElement('pubStmt')
        unpub: MeiElement = pubStmt.appendSubElement('unpub')
        unpub.text = (
            '''This MEI file was created by converter21's MEI writer. When
                   published, this unpub element should be removed, and the
                   enclosing pubStmt element should be properly filled out.'''
        )

        # sourceDesc: There are potentially multiple sources here, depending on what
        # metadata items we find.  One for the digital source, one for the published
        # printed source, one for a recorded source album (with maybe one for the
        # recorded source album track), and one for an unpublished manuscript source.
        # if sourceDesc ends up empty, we will not add it, so create separately and
        # only append if not empty.
        sourceDesc: MeiElement = MeiElement('sourceDesc')

        digitalSource: MeiElement | None = self.makeDigitalSource()
        if digitalSource is not None:
            sourceDesc.subElements.append(digitalSource)

        publishedSource: MeiElement | None = self.makePrintedSource()
        if publishedSource is not None:
            sourceDesc.subElements.append(publishedSource)

        if digitalSource is not None and publishedSource is not None:
            digitalSourceBibl: MeiElement | None = (
                digitalSource.findFirst('bibl', recurse=False)
            )
            publishedSourceBibl: MeiElement | None = (
                publishedSource.findFirst('bibl', recurse=False)
            )
            # relate them to eachother with 'otherFormat'.  They are the same,
            # just in different formats (printed vs digital).
            if digitalSourceBibl is not None and publishedSourceBibl is not None:
                digitalSourceBibl.attrib['xml:id'] = 'source0_digital'
                publishedSourceBibl.attrib['xml:id'] = 'source1_printed'
                digitalSourceBibl.appendSubElement(
                    'relatedItem',
                    {
                        'rel': 'otherFormat',
                        'target': '#source1_printed'
                    }
                )
                publishedSourceBibl.appendSubElement(
                    'relatedItem',
                    {
                        'rel': 'otherFormat',
                        'target': '#source0_digital'
                    }
                )

        recordedSource: MeiElement | None = self.makeRecordedSource()
        if recordedSource is not None:
            sourceDesc.subElements.append(recordedSource)

        unpublishedSource: MeiElement | None = self.makeUnpublishedSource()
        if unpublishedSource is not None:
            sourceDesc.subElements.append(unpublishedSource)

        if not sourceDesc.isEmpty():
            fileDesc.subElements.append(sourceDesc)

        return fileDesc

    def makeDigitalSource(self) -> MeiElement | None:
        if not self.anyExist(
                'EED', 'ENC', 'EEV', 'EFL', 'YEP', 'YER',
                'END', 'YEC', 'YEM', 'YEN', 'TXL', 'ONB'):
            return None

        source: MeiElement = MeiElement('source', {'type': 'digital'})
        bibl: MeiElement = source.appendSubElement('bibl')
        if self.simpleTitleElement is None:
            self.simpleTitleElement = self.makeSimpleTitleElement()
        if self.simpleTitleElement is not None:
            bibl.subElements.append(self.simpleTitleElement)
        if not self.simpleComposerElements:
            self.simpleComposerElements = self.makeSimpleComposerElements()
        if self.simpleComposerElements:
            bibl.subElements.extend(self.simpleComposerElements)

        editors: list[MeiMetadataItem] = self.contents.get('EED', [])
        encoders: list[MeiMetadataItem] = self.contents.get('ENC', [])
        versions: list[MeiMetadataItem] = self.contents.get('EEV', [])
        fileNumbers: list[MeiMetadataItem] = self.contents.get('EFL', [])
        publishers: list[MeiMetadataItem] = self.contents.get('YEP', [])
        releaseDates: list[MeiMetadataItem] = self.contents.get('YER', [])
        encodingDates: list[MeiMetadataItem] = self.contents.get('END', [])
        copyrights: list[MeiMetadataItem] = self.contents.get('YEC', [])
        copyrightStatements: list[MeiMetadataItem] = self.contents.get('YEM', [])
        copyrightCountries: list[MeiMetadataItem] = self.contents.get('YEN', [])
        textLanguages: list[MeiMetadataItem] = self.contents.get('TXL', [])
        notes: list[MeiMetadataItem] = self.contents.get('ONB', [])

        for editor in editors:
            editorEl: MeiElement = bibl.appendSubElement(
                'editor',
                {
                    'analog': 'humdrum:EED'
                }
            )
            editorEl.text = editor.meiValue

        if encoders:
            respStmt: MeiElement = bibl.appendSubElement('respStmt')

            for encoder in encoders:
                respEl: MeiElement = respStmt.appendSubElement('resp')
                respEl.text = 'encoder'
                persNameEl: MeiElement = respStmt.appendSubElement(
                    'persName',
                    {
                        'analog': 'humdrum:ENC'
                    }
                )
                persNameEl.text = encoder.meiValue

        for version in versions:
            versionEl: MeiElement = bibl.appendSubElement(
                'edition',
                {
                    'type': 'version',
                    'analog': 'humdrum:EEV'
                }
            )
            versionEl.text = version.meiValue

        for fileNumber in fileNumbers:
            fileNumberEl: MeiElement = bibl.appendSubElement(
                'extent',
                {
                    'type': 'fileNumber',
                    'unit': 'file',
                    'analog': 'humdrum:EFL'
                }
            )
            fileNumberEl.text = fileNumber.meiValue

        if publishers or releaseDates or encodingDates:
            imprint: MeiElement = bibl.appendSubElement('imprint')
            for publisher in publishers:
                publisherEl: MeiElement = imprint.appendSubElement(
                    'publisher',
                    {
                        'analog': 'humdrum:YEP'
                    }
                )
                publisherEl.text = publisher.meiValue

            for releaseDate in releaseDates:
                releaseDateEl: MeiElement = imprint.appendSubElement(
                    'date',
                    {
                        'type': 'releaseDate',
                        'analog': 'humdrum:YER'
                    }
                )
                releaseDateEl.fillInIsodate(releaseDate.value)
                releaseDateEl.text = releaseDate.meiValue

            for encodingDate in encodingDates:
                encodingDateEl: MeiElement = imprint.appendSubElement(
                    'date',
                    {
                        'type': 'encodingDate',
                        'analog': 'humdrum:END'
                    }
                )
                encodingDateEl.fillInIsodate(encodingDate.value)
                encodingDateEl.text = encodingDate.meiValue

        if copyrights or copyrightStatements or copyrightCountries:
            availability: MeiElement = bibl.appendSubElement('availability')
            for cpyright in copyrights:
                copyrightEl: MeiElement = availability.appendSubElement(
                    'useRestrict',
                    {
                        'type': 'copyright',
                        'analog': 'humdrum:YEC'
                    }
                )
                copyrightEl.text = cpyright.meiValue

            for copyrightStatement in copyrightStatements:
                copyrightStatementEl: MeiElement = availability.appendSubElement(
                    'useRestrict',
                    {
                        'type': 'copyrightStatement',
                        'analog': 'humdrum:YEM'
                    }
                )
                copyrightStatementEl.text = copyrightStatement.meiValue

            for copyrightCountry in copyrightCountries:
                copyrightCountryEl: MeiElement = availability.appendSubElement(
                    'useRestrict',
                    {
                        'type': 'copyrightCountry',
                        'analog': 'humdrum:YEN'
                    }
                )
                copyrightCountryEl.text = copyrightCountry.meiValue

        if notes:
            annot: MeiElement = bibl.appendSubElement('annot')
            language, allTheSameLanguage = self.getTextListLanguage(notes)
            lineGroup: MeiElement = annot.appendSubElement('lg')
            if allTheSameLanguage and language:
                lineGroup.attrib['xml:lang'] = language

            for note in notes:
                # <l> does not take @analog, so use @type instead (says Perry)
                line = lineGroup.appendSubElement('l', {'type': 'humdrum:ONB'})
                line.text = note.meiValue
                if note.value.language and not allTheSameLanguage:
                    line.attrib['xml:lang'] = note.value.language.lower()

        for textLanguage in textLanguages:
            textLanguageEl: MeiElement = bibl.appendSubElement(
                'textLang',
                {
                    'analog': 'humdrum:TXL'
                }
            )
            textLanguageEl.text = textLanguage.meiValue

        if bibl.isEmpty():
            # we check bibl because source is not empty: it contains bibl.
            return None
        return source

    def makePrintedSource(self) -> MeiElement | None:
        if not self.anyExist('LAR', 'PED', 'LOR', 'TRN', 'OCL', 'OVM', 'PTL',
                'PPR', 'PDT', 'PPP', 'PC#', 'mei:printedSourceCopyright'):
            return None

        arrangers: list[MeiMetadataItem] = self.contents.get('LAR', [])
        editors: list[MeiMetadataItem] = self.contents.get('PED', [])
        orchestrators: list[MeiMetadataItem] = self.contents.get('LOR', [])
        translators: list[MeiMetadataItem] = self.contents.get('TRN', [])
        collectors: list[MeiMetadataItem] = self.contents.get('OCL', [])
        volumeNumbers: list[MeiMetadataItem] = self.contents.get('OVM', [])
        volumeNames: list[MeiMetadataItem] = self.contents.get('PTL', [])
        publishers: list[MeiMetadataItem] = self.contents.get('PPR', [])
        datesPublished: list[MeiMetadataItem] = self.contents.get('PDT', [])
        locationsPublished: list[MeiMetadataItem] = self.contents.get('PPP', [])
        publisherCatalogNumbers: list[MeiMetadataItem] = self.contents.get('PC#', [])
        printedSourceCopyrights: list[MeiMetadataItem] = (
            self.contents.get('mei:printedSourceCopyright', [])
        )

        source: MeiElement = MeiElement('source', {'type': 'printed'})
        bibl: MeiElement = source.appendSubElement('bibl')

        for publisherCatalogNumber in publisherCatalogNumbers:
            identifierEl: MeiElement = bibl.appendSubElement(
                'identifier',
                {
                    'type': 'catalogNumber',
                    'analog': 'humdrum:PC#'
                }
            )
            identifierEl.text = publisherCatalogNumber.meiValue

        if self.simpleTitleElement is None:
            self.simpleTitleElement = self.makeSimpleTitleElement()
        if self.simpleTitleElement is not None:
            bibl.subElements.append(self.simpleTitleElement)
        if not self.simpleComposerElements:
            self.simpleComposerElements = self.makeSimpleComposerElements()
        if self.simpleComposerElements:
            bibl.subElements.extend(self.simpleComposerElements)

        for editor in editors:
            editorEl: MeiElement = bibl.appendSubElement(
                'editor',
                {
                    'analog': 'humdrum:PED'
                }
            )
            editorEl.text = editor.meiValue

        if arrangers or orchestrators or translators or collectors:
            # arrangers could technically go outside <respStmt>, but
            # Perry requests that they go inside <respStmt> for ease
            # of conversion to his proposed v. 6.
            respStmt: MeiElement = bibl.appendSubElement('respStmt')

            for arranger in arrangers:
                respEl: MeiElement = respStmt.appendSubElement('resp')
                respEl.text = 'arranger'
                arrangerEl: MeiElement = respStmt.appendSubElement(
                    'persName',
                    {
                        'analog': 'humdrum:LAR'
                    }
                )
                arrangerEl.text = arranger.meiValue

            for orchestrator in orchestrators:
                respEl = respStmt.appendSubElement('resp')
                respEl.text = 'orchestrator'
                persNameEl: MeiElement = respStmt.appendSubElement(
                    'persName',
                    {
                        'analog': 'humdrum:LOR'
                    }
                )
                persNameEl.text = orchestrator.meiValue

            for translator in translators:
                respEl = respStmt.appendSubElement('resp')
                respEl.text = 'translator'
                persNameEl = respStmt.appendSubElement(
                    'persName',
                    {
                        'analog': 'humdrum:TRN'
                    }
                )
                persNameEl.text = translator.meiValue

            for collector in collectors:
                respEl = respStmt.appendSubElement('resp')
                respEl.text = 'collector/transcriber'
                persNameEl = respStmt.appendSubElement(
                    'name',
                    {
                        'analog': 'humdrum:OCL'
                    }
                )
                persNameEl.text = collector.meiValue

        if publishers or datesPublished or locationsPublished:
            imprint: MeiElement = bibl.appendSubElement('imprint')
            for publisher in publishers:
                publisherEl: MeiElement = imprint.appendSubElement(
                    'publisher',
                    {
                        'analog': 'humdrum:PPR'
                    }
                )
                publisherEl.text = publisher.meiValue
            for datePublished in datesPublished:
                dateEl: MeiElement = imprint.appendSubElement(
                    'date',
                    {
                        'type': 'datePublished',
                        'analog': 'humdrum:PDT'
                    }
                )
                dateEl.text = datePublished.meiValue

            for locationPublished in locationsPublished:
                geogNameEl: MeiElement = imprint.appendSubElement(
                    'geogName',
                    {
                        'role': 'locationPublished',
                        'analog': 'humdrum:PPP'
                    }
                )
                geogNameEl.text = locationPublished.meiValue

        if printedSourceCopyrights:
            availability: MeiElement = bibl.appendSubElement('availability')
            for printedSourceCopyright in printedSourceCopyrights:
                useRestrict: MeiElement = availability.appendSubElement(
                    'useRestrict',
                    {
                        'type': 'copyright'
                    }
                )
                useRestrict.text = printedSourceCopyright.meiValue

        for volumeName, volumeNumber in zip(volumeNames, volumeNumbers):
            relatedItem: MeiElement = bibl.appendSubElement(
                'relatedItem',
                {
                    'rel': 'host'
                }
            )
            relBibl: MeiElement = relatedItem.appendSubElement('bibl')
            titleElement = relBibl.appendSubElement(
                'title',
                {
                    'analog': 'humdrum:PTL'
                }
            )
            titleElement.text = volumeName.meiValue

            biblScope: MeiElement = relBibl.appendSubElement(
                'biblScope',
                {
                    'analog': 'humdrum:OVM'
                }
            )
            biblScope.text = volumeNumber.meiValue

        if len(volumeNames) - len(volumeNumbers) > 0:
            # we ignore any extra volume numbers, since a number without a name
            # isn't interesting.
            for volumeName in volumeNames[len(volumeNumbers):]:
                relatedItem = bibl.appendSubElement(
                    'relatedItem',
                    {
                        'rel': 'host'
                    }
                )
                relBibl = relatedItem.appendSubElement('bibl')
                titleElement = relBibl.appendSubElement(
                    'title',
                    {
                        'analog': 'humdrum:PTL'
                    }
                )
                titleElement.text = volumeName.meiValue

        if bibl.isEmpty():
            return None
        return source

    def makeRecordedSource(self) -> MeiElement | None:
        if not self.anyExist('RTL', 'RC#', 'MGN', 'MPN', 'MPS', 'RNP',
                'MCN', 'RMM', 'RRD', 'RLC', 'RDT', 'MLC'):
            return None

        source: MeiElement = MeiElement('source', {'type': 'recording'})
        biblStruct: MeiElement = source.appendSubElement('biblStruct')

        albumTitles: list[MeiMetadataItem] = self.contents.get('RTL', [])
        albumCatalogNumbers: list[MeiMetadataItem] = self.contents.get('RC#', [])
        ensembleNames: list[MeiMetadataItem] = self.contents.get('MGN', [])
        performerNames: list[MeiMetadataItem] = self.contents.get('MPN', [])
        suspectedPerformerNames: list[MeiMetadataItem] = self.contents.get('MPS', [])
        producers: list[MeiMetadataItem] = self.contents.get('RNP', [])
        conductors: list[MeiMetadataItem] = self.contents.get('MCN', [])
        manufacturers: list[MeiMetadataItem] = self.contents.get('RMM', [])
        releaseDates: list[MeiMetadataItem] = self.contents.get('RRD', [])
        recordingLocations: list[MeiMetadataItem] = self.contents.get('RLC', [])
        recordingDates: list[MeiMetadataItem] = self.contents.get('RDT', [])
        trackNumbers: list[MeiMetadataItem] = self.contents.get('RT#', [])
        longestLen: int = 0
        longestLen = max(
            longestLen,
            len(albumTitles),
            len(albumCatalogNumbers),
            len(ensembleNames),
            len(performerNames),
            len(suspectedPerformerNames),
            len(producers),
            len(conductors),
            len(manufacturers),
            len(releaseDates),
            len(recordingLocations),
            len(recordingDates),
            len(trackNumbers),
        )

        for i in range(0, longestLen):
            if i < len(trackNumbers):
                analytic: MeiElement = biblStruct.appendSubElement('analytic')
                if self.simpleTitleElement is None:
                    self.simpleTitleElement = self.makeSimpleTitleElement()
                if self.simpleTitleElement is not None:
                    analytic.subElements.append(self.simpleTitleElement)
                biblScope = analytic.appendSubElement(
                    'biblScope',
                    {
                        'type': 'trackNumber',
                        'unit': 'track',
                        'analog': 'humdrum:RT#'
                    }
                )
                biblScope.text = trackNumbers[i].meiValue

            if (i < len(albumTitles)
                    or i < len(albumCatalogNumbers)
                    or i < len(ensembleNames)
                    or i < len(performerNames)
                    or i < len(suspectedPerformerNames)
                    or i < len(producers)
                    or i < len(conductors)
                    or i < len(manufacturers)
                    or i < len(releaseDates)
                    or i < len(recordingLocations)
                    or i < len(recordingDates)
                    or i < len(recordingDates)):
                monogr: MeiElement = biblStruct.appendSubElement('monogr')

                if i < len(albumTitles):
                    albumTitle: MeiMetadataItem = albumTitles[i]

                    albumTitleEl: MeiElement = monogr.appendSubElement(
                        'title',
                        {
                            'analog': 'humdrum:RTL'
                        }
                    )
                    albumTitleEl.text = albumTitle.meiValue

                if i < len(albumCatalogNumbers):
                    albumCatalogNumber: MeiMetadataItem = albumCatalogNumbers[i]
                    albumCatalogNumberEl: MeiElement = monogr.appendSubElement(
                        'identifier',
                        {
                            'type': 'albumCatalogNumber',
                            'analog': 'humdrum:RC#'
                        }
                    )
                    albumCatalogNumberEl.text = albumCatalogNumber.meiValue

                if (i < len(ensembleNames)
                        or i < len(performerNames)
                        or i < len(suspectedPerformerNames)
                        or i < len(producers)
                        or i < len(conductors)):
                    respStmt: MeiElement = monogr.appendSubElement('respStmt')

                    if i < len(ensembleNames):
                        ensembleName: MeiMetadataItem = ensembleNames[i]
                        respEl = respStmt.appendSubElement('resp')
                        respEl.text = 'performer'
                        ensembleNameEl: MeiElement = respStmt.appendSubElement(
                            'corpName',
                            {
                                'type': 'ensembleName',
                                'analog': 'humdrum:MGN'
                            }
                        )
                        ensembleNameEl.text = ensembleName.meiValue

                    if i < len(performerNames):
                        performerName: MeiMetadataItem = performerNames[i]
                        respEl = respStmt.appendSubElement('resp')
                        respEl.text = 'performer'
                        performerNameEl: MeiElement = respStmt.appendSubElement(
                            'persName',
                            {
                                'analog': 'humdrum:MPN'
                            }
                        )
                        performerNameEl.text = performerName.meiValue

                    if i < len(suspectedPerformerNames):
                        suspectedPerformerName: MeiMetadataItem = suspectedPerformerNames[i]
                        respEl = respStmt.appendSubElement('resp')
                        respEl.text = 'performer'
                        suspectedPerformerNameEl: MeiElement = respStmt.appendSubElement(
                            'persName',
                            {
                                'cert': 'medium',
                                'analog': 'humdrum:MPS'
                            }
                        )
                        suspectedPerformerNameEl.text = suspectedPerformerName.meiValue

                    if i < len(producers):
                        producer: MeiMetadataItem = producers[i]
                        respEl = respStmt.appendSubElement('resp')
                        respEl.text = 'producer'
                        persNameEl: MeiElement = respStmt.appendSubElement(
                            'name',
                            {
                                'analog': 'humdrum:RNP'
                            }
                        )
                        persNameEl.text = producer.meiValue

                    if i < len(conductors):
                        conductor: MeiMetadataItem = conductors[i]
                        respEl = respStmt.appendSubElement('resp')
                        respEl.text = 'conductor'
                        persNameEl = respStmt.appendSubElement(
                            'persName',
                            {
                                'analog': 'humdrum:MCN'
                            }
                        )
                        persNameEl.text = conductor.meiValue

                if (i < len(manufacturers)
                        or i < len(releaseDates)
                        or i < len(recordingLocations)
                        or i < len(recordingDates)):
                    imprint: MeiElement = monogr.appendSubElement('imprint')

                    if i < len(manufacturers):
                        manufacturer: MeiMetadataItem = manufacturers[i]
                        manufacturerEl: MeiElement = imprint.appendSubElement(
                            'corpName',
                            {
                                'role': 'production/distribution',
                                'analog': 'humdrum:RMM'
                            }
                        )
                        manufacturerEl.text = manufacturer.meiValue

                    if i < len(releaseDates):
                        releaseDate: MeiMetadataItem = releaseDates[i]
                        releaseDateEl: MeiElement = imprint.appendSubElement(
                            'date',
                            {
                                'type': 'releaseDate',
                                'analog': 'humdrum:RRD'
                            }
                        )
                        releaseDateEl.fillInIsodate(releaseDate.value)
                        releaseDateEl.text = releaseDate.meiValue

                    if i < len(recordingLocations):
                        recordingLocation: MeiMetadataItem = recordingLocations[i]
                        recordingLocationEl: MeiElement = imprint.appendSubElement(
                            'geogName',
                            {
                                'role': 'recordingLocation',
                                'analog': 'humdrum:RLC'
                            }
                        )
                        recordingLocationEl.text = recordingLocation.meiValue

                    if i < len(recordingDates):
                        recordingDate: MeiMetadataItem = recordingDates[i]
                        recordingDateEl: MeiElement = imprint.appendSubElement(
                            'date',
                            {
                                'type': 'recordingDate',
                                'analog': 'humdrum:RDT'
                            }
                        )
                        recordingDateEl.fillInIsodate(recordingDate.value)
                        recordingDateEl.text = recordingDate.meiValue

        if biblStruct.isEmpty():
            return None
        return source

    def makeUnpublishedSource(self) -> MeiElement | None:
        if not self.anyExist('SMS', 'YOR', 'SML', 'YOO', 'YOE', 'YOY', 'SMA'):
            return None

        source: MeiElement = MeiElement('source', {'type': 'unpub'})
        bibl: MeiElement = source.appendSubElement('bibl')

        manuscriptNames: list[MeiMetadataItem] = self.contents.get('SMS', [])
        moreManuscriptNames: list[MeiMetadataItem] = self.contents.get('YOR', [])
        manuscriptLocations: list[MeiMetadataItem] = self.contents.get('SML', [])
        manuscriptOwners: list[MeiMetadataItem] = self.contents.get('YOO', [])
        editors: list[MeiMetadataItem] = self.contents.get('YOE', [])
        copyrightDates: list[MeiMetadataItem] = self.contents.get('YOY', [])
        acknowledgments: list[MeiMetadataItem] = self.contents.get('SMA', [])

        for manuscriptName in manuscriptNames:
            manuscriptNameEl: MeiElement = bibl.appendSubElement(
                'identifier',
                {
                    'analog': 'humdrum:SMS'
                }
            )
            manuscriptNameEl.text = manuscriptName.meiValue

        for manuscriptName in moreManuscriptNames:
            manuscriptNameEl = bibl.appendSubElement(
                'identifier',
                {
                    'analog': 'humdrum:YOR'
                }
            )
            manuscriptNameEl.text = manuscriptName.meiValue

        # do both again as <title>
        for manuscriptName in manuscriptNames:
            manuscriptNameEl = bibl.appendSubElement(
                'title',
                {
                    'analog': 'humdrum:SMS'
                }
            )
            manuscriptNameEl.text = manuscriptName.meiValue

        for manuscriptName in moreManuscriptNames:
            manuscriptNameEl = bibl.appendSubElement(
                'title',
                {
                    'analog': 'humdrum:YOR'
                }
            )
            manuscriptNameEl.text = manuscriptName.meiValue

        for manuscriptLocation in manuscriptLocations:
            manuscriptLocationEl = bibl.appendSubElement(
                'repository',
                {
                    'analog': 'humdrum:SML'
                }
            )
            manuscriptLocationEl.text = manuscriptLocation.meiValue

        for manuscriptOwner in manuscriptOwners:
            manuscriptOwnerEl = bibl.appendSubElement(
                'name',
                {
                    'role': 'manuscriptOwner',
                    'analog': 'humdrum:YOO'
                }
            )
            manuscriptOwnerEl.text = manuscriptOwner.meiValue

        for editor in editors:
            editorEl = bibl.appendSubElement(
                'editor',
                {
                    'analog': 'humdrum:YOE'
                }
            )
            editorEl.text = editor.meiValue

        for copyrightDate in copyrightDates:
            copyrightDateEl = bibl.appendSubElement(
                'date',
                {
                    'type': 'copyrightDate',
                    'analog': 'humdrum:YOY'
                }
            )
            copyrightDateEl.fillInIsodate(copyrightDate.value)
            copyrightDateEl.text = copyrightDate.meiValue

        if acknowledgments:
            annot: MeiElement = bibl.appendSubElement(
                'annot',
                {
                    'type': 'manuscriptAccessAcknowledgment',
                }
            )
            language, allTheSameLanguage = self.getTextListLanguage(acknowledgments)
            lineGroup: MeiElement = annot.appendSubElement('lg')
            if allTheSameLanguage and language:
                lineGroup.attrib['xml:lang'] = language

        for acknowledgment in acknowledgments:
            # <l> does not take @analog, so use @type instead (says Perry)
            line = lineGroup.appendSubElement('l', {'type': 'humdrum:SMA'})
            line.text = acknowledgment.meiValue
            if acknowledgment.value.language and not allTheSameLanguage:
                line.attrib['xml:lang'] = acknowledgment.value.language.lower()

        if bibl.isEmpty():
            return None
        return source

    def makeSimpleComposerElements(self) -> list[MeiElement]:
        # Just all the COMs.  If no COMs, then do the COCs (corporate), the COAs (attributed),
        # the COSs (suspected), or the COLs (aliases), in that order of preference.
        output: list[MeiElement] = []
        composers: list[MeiMetadataItem] = self.contents.get('COM', [])
        if not composers:
            composers = self.contents.get('COC', [])
        if not composers:
            composers = self.contents.get('COA', [])
        if not composers:
            composers = self.contents.get('COS', [])
        if not composers:
            composers = self.contents.get('COL', [])
        if not composers:
            return output

        for composer in composers:
            composerElement = MeiElement('composer')
            output.append(composerElement)

            # composer cert
            composerCert: str = ''
            if composer.value.role == 'attributedComposer':
                composerCert = 'medium'
            elif composer.value.role == 'suspectedComposer':
                composerCert = 'low'
            if composerCert:
                # We put @cert on <composer>, not on <persName>.
                # The uncertainty is about whether this composer was involved,
                # not about what this composer's name was.
                composerElement.attrib['cert'] = composerCert

            nameElement: MeiElement = composerElement.appendSubElement('persName')
            nameElement.text = composer.meiValue

            # adjust nameElement.name and nameElement.attrib as necessary
            if composer.value.role == 'composerCorporate':
                nameElement.name = 'corpName'
            elif composer.value.role == 'composerAlias':
                nameElement.attrib['type'] = 'alias'

        return output

    def makeComposerElements(self) -> list[MeiElement]:
        # returns a list of <composer> elements
        # As a side effect, generates any <mads> XML for composer personal info
        # (e.g. birth and death dates/places, nationality), stashing them in
        # self.madsElements to emit later.

        output: list[MeiElement] = []
        composers: list[MeiMetadataItem] = self.contents.get('COM', [])
        attributedComposers: list[MeiMetadataItem] = self.contents.get('COA', [])
        suspectedComposers: list[MeiMetadataItem] = self.contents.get('COS', [])
        corporateComposers: list[MeiMetadataItem] = self.contents.get('COC', [])
        composerAliases: list[MeiMetadataItem] = self.contents.get('COL', [])
        composerDates: list[MeiMetadataItem] = self.contents.get('CDT', [])
        composerBirthPlaces: list[MeiMetadataItem] = self.contents.get('CBL', [])
        composerDeathPlaces: list[MeiMetadataItem] = self.contents.get('CDL', [])
        composerNationalities: list[MeiMetadataItem] = self.contents.get('CNT', [])
        # TODO: Prefer composer birth/death dates from Contributor value.
        # TODO: But first make Humdrum importer put them there.
        allComposers: list[MeiMetadataItem] = (
            composers + attributedComposers + suspectedComposers + corporateComposers
        )

        madsXmlIdIndex: int = 0
        for i, composer in enumerate(allComposers):
            # Assume association is done by ordering of the metadata item arrays
            # Currently our Humdrum importer assumes this ordering should match
            # the ordering in the file.
            composerAlias: MeiMetadataItem | None = None
            composerBirthAndDeathDate: MeiMetadataItem | None = None
            composerBirthPlace: MeiMetadataItem | None = None
            composerDeathPlace: MeiMetadataItem | None = None
            composerNationality: MeiMetadataItem | None = None
            if i < len(composerAliases):
                composerAlias = composerAliases[i]
            if i < len(composerDates):
                composerBirthAndDeathDate = composerDates[i]
            if i < len(composerBirthPlaces):
                composerBirthPlace = composerBirthPlaces[i]
            if i < len(composerDeathPlaces):
                composerDeathPlace = composerDeathPlaces[i]
            if i < len(composerNationalities):
                composerNationality = composerNationalities[i]

            composerElement = MeiElement('composer')
            output.append(composerElement)

            # composer persName ('COM'/'COA'/'COS')
            composerAnalog: str = 'humdrum:COM'
            composerCert: str = ''
            composerNameElementName: str = 'persName'
            if composer.value.role == 'attributedComposer':
                composerAnalog = 'humdrum:COA'
                composerCert = 'medium'
            elif composer.value.role == 'suspectedComposer':
                composerAnalog = 'humdrum:COS'
                composerCert = 'low'
            elif composer.value.role == 'composerCorporate':
                composerAnalog = 'humdrum:COC'
                composerNameElementName = 'corpName'

            if composerCert:
                # We put @cert on <composer>, not on <persName>.
                # The uncertainty is about whether this composer was involved,
                # not about this composer's name.
                composerElement.attrib['cert'] = composerCert

            nameElement: MeiElement = composerElement.appendSubElement(
                composerNameElementName,
                {
                    'analog': composerAnalog
                }
            )
            nameElement.text = composer.meiValue

            # MADS-style authority records (personal info about a composer)
            if (composerBirthAndDeathDate is None
                    and composerBirthPlace is None
                    and composerDeathPlace is None
                    and composerNationality is None):
                # nothing for MADS
                continue

            madsXmlId: str = f'mads{madsXmlIdIndex}'
            madsXmlIdIndex += 1

            # reference <mads> element from composer's nameElement
            nameElement.attrib['auth.uri'] = '#' + madsXmlId

            # There is extra info about the composer, that will need to go
            # in <extMeta><madsCollection><mads>
            if self.madsCollection is None:
                schemaLoc: str = (
                    'http://www.loc.gov/mads/v2 https://www.loc.gov/standards/mads/mads-2-1.xsd'
                )
                self.madsCollection = MeiElement(
                    'madsCollection',
                    {
                        'xmlns:xsi': 'http://www.w3.org/2001/XMLSchema-instance',
                        'xsi:schemaLocation': schemaLoc,
                        'xmlns': 'http://www.loc.gov/mads/v2'
                    }
                )

            mads = self.madsCollection.appendSubElement('mads', {'ID': madsXmlId})

            authority: MeiElement = mads.appendSubElement('authority')
            name: MeiElement = authority.appendSubElement('name')
            if composerNameElementName == 'corpName':
                name.attrib['type'] = 'corporate'
            namePart: MeiElement = name.appendSubElement('namePart')
            namePart.text = composer.meiValue

            # composerAlias ('humdrum:COL') goes in <mads><variant>
            if composerAlias:
                variant: MeiElement = mads.appendSubElement(
                    'variant',
                    {
                        'type': 'other',
                        'otherType': 'humdrum:COL'
                    }
                )
                name = variant.appendSubElement('name')
                namePart = name.appendSubElement('namePart')
                namePart.text = composerAlias.meiValue

            # extra info goes in <mads><personInfo>
            personInfo: MeiElement = mads.appendSubElement('personInfo')
            if composerBirthAndDeathDate is not None:
                isodate: str = (
                    M21Utilities.isoDateFromM21DatePrimitive(
                        composerBirthAndDeathDate.value,
                        returnEDTFString=True
                    )
                )
                isodateBirth: str = ''
                isodateDeath: str = ''
                isodateBirth, isodateDeath = isodate.split('/')
                birthDate: MeiElement = personInfo.appendSubElement(
                    'birthDate',
                    {
                        'encoding': 'edtf'
                    }
                )
                birthDate.text = isodateBirth

                if isodateDeath:
                    deathDate: MeiElement = personInfo.appendSubElement(
                        'deathDate',
                        {
                            'encoding': 'edtf'
                        }
                    )
                    deathDate.text = isodateDeath

            if composerBirthPlace is not None:
                birthPlace: MeiElement = personInfo.appendSubElement('birthPlace')
                birthPlace.text = composerBirthPlace.meiValue

            if composerDeathPlace is not None:
                deathPlace: MeiElement = personInfo.appendSubElement('deathPlace')
                deathPlace.text = composerDeathPlace.meiValue

            if composerNationality is not None:
                nationality: MeiElement = personInfo.appendSubElement('nationality')
                nationality.text = composerNationality.meiValue

        return output

    @staticmethod
    def getBestName(
        mmItems: list[MeiMetadataItem],
        requiredLanguage: str = ''
    ) -> MeiMetadataItem | None:
        # If requiredLanguage is set, return the first item with that language,
        # otherwise return the first item that is untranslated.
        # If nothing is found that has the right language/is untranslated,
        # just return the first item (or None, if mmItems is empty).
        bestName: MeiMetadataItem | None = None
        for mmItem in mmItems:
            if not isinstance(mmItem.value, m21.metadata.Text):
                continue

            if requiredLanguage:
                if mmItem.value.language.lower() == requiredLanguage.lower():
                    bestName = mmItem
                    break
            else:
                if mmItem.value.isTranslated is not True:
                    bestName = mmItem
                    break

        if bestName is None and mmItems:
            bestName = mmItems[0]
        return bestName

    def makeSimpleTitleElement(self) -> MeiElement:
        # first untranslated OTL (or just first OTL, if all are translated)
        # plus first OMD in that same language goes in the <title> element
        # (no titlePart).
        # If there are no OTLs/OMDs, empty <title/> is returned
        # Note: no @analog, the individual titles will have @analog elsewhere.
        titles: list[MeiMetadataItem] = self.contents.get('OTL', [])
        movementNames: list[MeiMetadataItem] = self.contents.get('OMD', [])
        titleElement = MeiElement('title')
        firstLang: str = ''
        bestTitle: MeiMetadataItem | None = self.getBestName(titles)
        if bestTitle is not None and bestTitle.value.language:
            firstLang = bestTitle.value.language
        bestMovementName: MeiMetadataItem | None = self.getBestName(movementNames, firstLang)
        if (bestTitle is None
                and bestMovementName is not None
                and bestMovementName.value.language):
            firstLang = bestMovementName.value.language.lower()

        if bestTitle is None and bestMovementName is None:
            return titleElement

        if firstLang:
            titleElement.attrib['xml:lang'] = firstLang

        if bestTitle and bestMovementName:
            titleElement.text = bestTitle.meiValue + ', ' + bestMovementName.meiValue
        elif bestTitle:
            titleElement.text = bestTitle.meiValue
        elif bestMovementName:
            titleElement.text = bestMovementName.meiValue
        return titleElement

    def makeTitleElements(self) -> list[MeiElement]:
        mainTitles: list[MeiMetadataItem] = self.contents.get('OTL', [])
        alternativeTitles: list[MeiMetadataItem] = self.contents.get('OTA', [])
        popularTitles: list[MeiMetadataItem] = self.contents.get('OTP', [])
        plainNumbers: list[MeiMetadataItem] = self.contents.get('ONM', [])
        movementNumbers: list[MeiMetadataItem] = self.contents.get('OMV', [])
        movementNames: list[MeiMetadataItem] = self.contents.get('OMD', [])
        opusNumbers: list[MeiMetadataItem] = self.contents.get('OPS', [])
        actNumbers: list[MeiMetadataItem] = self.contents.get('OAC', [])
        sceneNumbers: list[MeiMetadataItem] = self.contents.get('OSC', [])

        untranslatedTitleElement = MeiElement('title')
        translatedTitleElement = MeiElement('title')
        alternativeTitleElements: list[MeiElement] = []
        popularTitleElements: list[MeiElement] = []

        # First all the main titles (OTL).
        titlePart: MeiElement
        for mainTitle in mainTitles:
            lang: str = ''
            if t.TYPE_CHECKING:
                assert isinstance(mainTitle.value, m21.metadata.Text)
            isTranslated = mainTitle.value.isTranslated is True
            if mainTitle.value.language:
                lang = mainTitle.value.language.lower()

            attrib = {}
            if isTranslated:
                attrib['type'] = 'translated'
            else:
                attrib['type'] = 'main'
            attrib['analog'] = 'humdrum:OTL'
            if lang:
                attrib['xml:lang'] = lang

            if isTranslated:
                titlePart = translatedTitleElement.appendSubElement('titlePart', attrib)
            else:
                titlePart = untranslatedTitleElement.appendSubElement('titlePart', attrib)
            titlePart.text = mainTitle.meiValue

        # Then any movement name(s) (OMD).
        for movementName in movementNames:
            lang = ''
            if t.TYPE_CHECKING:
                assert isinstance(movementName.value, m21.metadata.Text)
            isTranslated = movementName.value.isTranslated is True
            if movementName.value.language:
                lang = movementName.value.language.lower()

            attrib = {'type': 'movementName', 'analog': 'humdrum:OMD'}
            if lang:
                attrib['xml:lang'] = lang

            if isTranslated:
                titlePart = translatedTitleElement.appendSubElement('titlePart', attrib)
            else:
                titlePart = untranslatedTitleElement.appendSubElement('titlePart', attrib)
            titlePart.text = movementName.meiValue

        # Then any number(s) (ONM).
        for plainNumber in plainNumbers:
            lang = ''
            if t.TYPE_CHECKING:
                assert isinstance(plainNumber.value, m21.metadata.Text)
            isTranslated = plainNumber.value.isTranslated is True
            if plainNumber.value.language:
                lang = plainNumber.value.language.lower()

            attrib = {'type': 'number', 'analog': 'humdrum:ONM'}
            if lang:
                attrib['xml:lang'] = lang

            if isTranslated:
                titlePart = translatedTitleElement.appendSubElement('titlePart', attrib)
            else:
                titlePart = untranslatedTitleElement.appendSubElement('titlePart', attrib)
            titlePart.text = plainNumber.meiValue

        # Then any movement number(s) (OMV).
        for movementNumber in movementNumbers:
            lang = ''
            if t.TYPE_CHECKING:
                assert isinstance(movementNumber.value, m21.metadata.Text)
            isTranslated = movementNumber.value.isTranslated is True
            if movementNumber.value.language:
                lang = movementNumber.value.language.lower()

            attrib = {'type': 'movementNumber', 'analog': 'humdrum:OMV'}
            if lang:
                attrib['xml:lang'] = lang

            if isTranslated:
                titlePart = translatedTitleElement.appendSubElement('titlePart', attrib)
            else:
                titlePart = untranslatedTitleElement.appendSubElement('titlePart', attrib)
            titlePart.text = movementNumber.meiValue

        # Then any opus number(s) (OPS).
        for opusNumber in opusNumbers:
            lang = ''
            if t.TYPE_CHECKING:
                assert isinstance(opusNumber.value, m21.metadata.Text)
            isTranslated = opusNumber.value.isTranslated is True
            if opusNumber.value.language:
                lang = opusNumber.value.language.lower()

            attrib = {'type': 'opusNumber', 'analog': 'humdrum:OPS'}
            if lang:
                attrib['xml:lang'] = lang

            if isTranslated:
                titlePart = translatedTitleElement.appendSubElement('titlePart', attrib)
            else:
                titlePart = untranslatedTitleElement.appendSubElement('titlePart', attrib)
            titlePart.text = opusNumber.meiValue

        # Then any act number(s) (OAC).
        for actNumber in actNumbers:
            lang = ''
            if t.TYPE_CHECKING:
                assert isinstance(actNumber.value, m21.metadata.Text)
            isTranslated = actNumber.value.isTranslated is True
            if actNumber.value.language:
                lang = actNumber.value.language.lower()

            attrib = {'type': 'actNumber', 'analog': 'humdrum:OAC'}
            if lang:
                attrib['xml:lang'] = lang

            if isTranslated:
                titlePart = translatedTitleElement.appendSubElement('titlePart', attrib)
            else:
                titlePart = untranslatedTitleElement.appendSubElement('titlePart', attrib)
            titlePart.text = actNumber.meiValue

        # Then any scene number(s) (OSC).
        for sceneNumber in sceneNumbers:
            lang = ''
            if t.TYPE_CHECKING:
                assert isinstance(sceneNumber.value, m21.metadata.Text)
            isTranslated = sceneNumber.value.isTranslated is True
            if sceneNumber.value.language:
                lang = sceneNumber.value.language.lower()

            attrib = {'type': 'sceneNumber', 'analog': 'humdrum:OSC'}
            if lang:
                attrib['xml:lang'] = lang

            if isTranslated:
                titlePart = translatedTitleElement.appendSubElement('titlePart', attrib)
            else:
                titlePart = untranslatedTitleElement.appendSubElement('titlePart', attrib)
            titlePart.text = sceneNumber.meiValue

        # Separately, any alternative titles (OTA) (no titleParts)
        for alternativeTitle in alternativeTitles:
            lang = ''
            if t.TYPE_CHECKING:
                assert isinstance(alternativeTitle.value, m21.metadata.Text)
            isTranslated = alternativeTitle.value.isTranslated is True
            if alternativeTitle.value.language:
                lang = alternativeTitle.value.language.lower()

            attrib = {'type': 'alternative', 'analog': 'humdrum:OTA'}
            if lang:
                attrib['xml:lang'] = lang

            alternativeTitleEl = MeiElement('title', attrib)
            alternativeTitleEl.text = alternativeTitle.meiValue
            alternativeTitleElements.append(alternativeTitleEl)

        # Separately, any popular titles (OTP) (no titleParts)
        for popularTitle in popularTitles:
            lang = ''
            if t.TYPE_CHECKING:
                assert isinstance(popularTitle.value, m21.metadata.Text)
            isTranslated = popularTitle.value.isTranslated is True
            if popularTitle.value.language:
                lang = popularTitle.value.language.lower()

            attrib = {'type': 'popular', 'analog': 'humdrum:OTP'}
            if lang:
                attrib['xml:lang'] = lang

            popularTitleEl = MeiElement('title', attrib)
            popularTitleEl.text = popularTitle.meiValue
            popularTitleElements.append(popularTitleEl)

        # roll them all up into a list
        titleElements: list[MeiElement] = []
        if not untranslatedTitleElement.isEmpty():
            untranslatedTitleElement.attrib['type'] = 'uniform'
            titleElements.append(untranslatedTitleElement)
        if not translatedTitleElement.isEmpty():
            translatedTitleElement.attrib['type'] = 'translated'
            titleElements.append(translatedTitleElement)

        if alternativeTitleElements:
            titleElements.extend(alternativeTitleElements)
        if popularTitleElements:
            titleElements.extend(popularTitleElements)

        return titleElements

    def makeEncodingDescElement(self) -> MeiElement | None:
        encodingNotes: list[MeiMetadataItem] = self.contents.get('RNB', [])
        encodingWarnings: list[MeiMetadataItem] = self.contents.get('RWB', [])
        softwares: list[MeiMetadataItem] = self.contents.get('software', [])

        converter21AlreadyThere: bool = False
        for software in softwares:
            if software.meiValue == SharedConstants._CONVERTER21_NAME:
                version = software.meiOtherAttribs.get('version', '')
                if version == SharedConstants._CONVERTER21_VERSION:
                    converter21AlreadyThere = True

        if not encodingNotes and not encodingWarnings and not converter21AlreadyThere:
            return None

        encodingDesc: MeiElement = MeiElement('encodingDesc')

        if encodingNotes or encodingWarnings:
            editorialDecl: MeiElement = encodingDesc.appendSubElement('editorialDecl')
            p: MeiElement = editorialDecl.appendSubElement('p')

            language: str | None
            allTheSameLanguage: bool
            line: MeiElement
            if encodingNotes:
                language, allTheSameLanguage = self.getTextListLanguage(encodingNotes)
                lineGroup: MeiElement = p.appendSubElement('lg')
                if allTheSameLanguage and language:
                    lineGroup.attrib['xml:lang'] = language
                for note in encodingNotes:
                    # <l> does not take @analog, so use @type instead (says Perry)
                    line = lineGroup.appendSubElement('l', {'type': 'humdrum:RNB'})
                    line.text = note.meiValue
                    if note.value.language and not allTheSameLanguage:
                        line.attrib['xml:lang'] = note.value.language.lower()

            if encodingWarnings:
                language, allTheSameLanguage = self.getTextListLanguage(encodingWarnings)
                lineGroup = p.appendSubElement('lg')
                if allTheSameLanguage and language:
                    lineGroup.attrib['xml:lang'] = language
                for warning in encodingWarnings:
                    # <l> does not take @analog, so use @type instead (says Perry)
                    line = lineGroup.appendSubElement('l', {'type': 'humdrum:RWB'})
                    line.text = warning.meiValue
                    if warning.value.language and not allTheSameLanguage:
                        line.attrib['xml:lang'] = warning.value.language.lower()

        if softwares or not converter21AlreadyThere:
            appInfo: MeiElement = encodingDesc.appendSubElement('appInfo')

            # add converter21 if it wasn't already there
            if not converter21AlreadyThere:
                application: MeiElement = appInfo.appendSubElement(
                    'application',
                    {
                        'version': SharedConstants._CONVERTER21_VERSION
                    }
                )
                name: MeiElement = application.appendSubElement('name')
                name.text = SharedConstants._CONVERTER21_NAME

            for software in softwares:
                application = appInfo.appendSubElement(
                    'application',
                    software.meiOtherAttribs
                )
                name = application.appendSubElement('name')
                name.text = software.meiValue

        return encodingDesc

    @staticmethod
    def getTextListLanguage(textItems: list[MeiMetadataItem]) -> tuple[str | None, bool]:
        # returns tuple(theLanguage: str | None, allTheSameLanguage: bool)
        theLanguage: str | None = None
        allTheSameLanguage: bool = True
        for textItem in textItems:
            if t.TYPE_CHECKING:
                assert isinstance(textItem.value, m21.metadata.Text)
            if theLanguage is None and textItem.value.language is not None:
                theLanguage = textItem.value.language.lower()
                continue
            if theLanguage is not None and textItem.value.language is not None:
                if theLanguage != textItem.value.language.lower():
                    allTheSameLanguage = False
                    break

        return theLanguage, allTheSameLanguage

    def makeWorkListElement(self) -> MeiElement | None:
        # the main (encoded) work
        catalogNumbers: list[MeiMetadataItem] = self.contents.get('SCA', [])
        catalogAbbrevNumbers: list[MeiMetadataItem] = self.contents.get('SCT', [])
        opusNumbers: list[MeiMetadataItem] = self.contents.get('OPS', [])
        creationDates: list[MeiMetadataItem] = self.contents.get('ODT', [])
        creationCountries: list[MeiMetadataItem] = self.contents.get('OCY', [])
        creationSettlements: list[MeiMetadataItem] = self.contents.get('OPC', [])
        creationRegions: list[MeiMetadataItem] = self.contents.get('ARE', [])
        creationLatLongs: list[MeiMetadataItem] = self.contents.get('ARL', [])
        lyricists: list[MeiMetadataItem] = self.contents.get('LYR', [])
        librettists: list[MeiMetadataItem] = self.contents.get('LIB', [])
        dedicatees: list[MeiMetadataItem] = self.contents.get('ODE', [])
        funders: list[MeiMetadataItem] = self.contents.get('OCO', [])
        languages: list[MeiMetadataItem] = self.contents.get('TXO', [])
        histories: list[MeiMetadataItem] = self.contents.get('HAO', [])
        instrumentLists: list[MeiMetadataItem] = self.contents.get('AIN', [])
        forms: list[MeiMetadataItem] = self.contents.get('AFR', [])
        genres: list[MeiMetadataItem] = self.contents.get('AGN', [])
        modes: list[MeiMetadataItem] = self.contents.get('AMD', [])
        meters: list[MeiMetadataItem] = self.contents.get('AMT', [])
        styles: list[MeiMetadataItem] = self.contents.get('AST', [])
        firstPerformanceDates: list[MeiMetadataItem] = self.contents.get('MPD', [])
        performanceDates: list[MeiMetadataItem] = self.contents.get('MDT', [])
        performanceDates.extend(self.contents.get('MRD', []))
        performanceLocations: list[MeiMetadataItem] = self.contents.get('MLC', [])

        # related works
        parentWorkTitles: list[MeiMetadataItem] = self.contents.get('OPR', [])
        groupWorkTitles: list[MeiMetadataItem] = self.contents.get('GTL', [])
        associatedWorkTitles: list[MeiMetadataItem] = self.contents.get('GAW', [])
        # GCO and ACO have the same definition: "Collection designation, such
        # as Norton Scores, Smithsonian Collection, etc."
        collectionWorkTitles: list[MeiMetadataItem] = self.contents.get('GCO', [])
        collectionWorkTitles.extend(self.contents.get('ACO', []))

        parentWorkXmlId: str = ''
        groupWorkXmlId: str = ''
        associatedWorkXmlId: str = ''
        collectionWorkXmlId: str = ''

        workList: MeiElement | None = None
        workNumber: int = 0

        # the parent work
        if parentWorkTitles:
            if workList is None:
                workList = MeiElement('workList')

            parentWorkXmlId = f'work{workNumber}_parent'
            parentWork = workList.appendSubElement(
                'work',
                {
                    'xml:id': parentWorkXmlId,
                    'type': 'parent'
                }
            )
            workNumber += 1

            for parentWorkTitle in parentWorkTitles:
                titleElement = parentWork.appendSubElement(
                    'title',
                    {
                        'analog': 'humdrum:OPR'
                    }
                )
                titleElement.text = parentWorkTitle.meiValue

        # the group work
        if groupWorkTitles:
            if workList is None:
                workList = MeiElement('workList')

            groupWorkXmlId = f'work{workNumber}_group'
            groupWork = workList.appendSubElement(
                'work',
                {
                    'xml:id': groupWorkXmlId,
                    'type': 'group'
                }
            )
            workNumber += 1

            for groupWorkTitle in groupWorkTitles:
                titleElement = groupWork.appendSubElement(
                    'title',
                    {
                        'analog': 'humdrum:GTL'
                    }
                )
                titleElement.text = groupWorkTitle.meiValue

        # the associated work
        if associatedWorkTitles:
            if workList is None:
                workList = MeiElement('workList')

            associatedWorkXmlId = f'work{workNumber}_associated'
            associatedWork = workList.appendSubElement(
                'work',
                {
                    'xml:id': associatedWorkXmlId,
                    'type': 'associated'
                }
            )
            workNumber += 1

            for associatedWorkTitle in associatedWorkTitles:
                titleElement = associatedWork.appendSubElement(
                    'title',
                    {
                        'analog': 'humdrum:GAW'
                    }
                )
                titleElement.text = associatedWorkTitle.meiValue

        # the collection work
        if collectionWorkTitles:
            if workList is None:
                workList = MeiElement('workList')

            collectionWorkXmlId = f'work{workNumber}_collection'
            collectionWork = workList.appendSubElement(
                'work',
                {
                    'xml:id': collectionWorkXmlId,
                    'type': 'collection'
                }
            )
            workNumber += 1

            for collectionWorkTitle in collectionWorkTitles:
                titleElement = collectionWork.appendSubElement(
                    'title',
                    {
                        'analog': 'humdrum:GCO'
                    }
                )
                titleElement.text = collectionWorkTitle.meiValue

        titleElements: list[MeiElement] = self.makeTitleElements()
        composerElements: list[MeiElement] = self.makeComposerElements()

        # the main (encoded) work
        if (catalogNumbers
                or catalogAbbrevNumbers
                or opusNumbers
                or titleElements
                or creationDates
                or creationCountries
                or creationSettlements
                or creationRegions
                or creationLatLongs
                or composerElements
                or lyricists
                or librettists
                or dedicatees
                or funders
                or languages
                or histories
                or instrumentLists
                or forms
                or genres
                or modes
                or meters
                or styles
                or firstPerformanceDates
                or performanceDates
                or performanceLocations):
            if workList is None:
                workList = MeiElement('workList')

            self.mainWorkXmlId = f'work{workNumber}_encoded'
            workNumber += 1

            theWork = workList.appendSubElement(
                'work',
                {
                    'xml:id': self.mainWorkXmlId,
                    'type': 'encoded'
                }
            )

            # <identifier>
            for catalogNumber in catalogNumbers:
                identifierElement = theWork.appendSubElement(
                    'identifier',
                    {
                        'analog': 'humdrum:SCA'
                    }
                )
                identifierElement.text = catalogNumber.meiValue

            for catalogAbbrevNumber in catalogAbbrevNumbers:
                identifierElement = theWork.appendSubElement(
                    'identifier',
                    {
                        'analog': 'humdrum:SCT'
                    }
                )
                identifierElement.text = catalogAbbrevNumber.meiValue

            for opusNumber in opusNumbers:
                identifierElement = theWork.appendSubElement(
                    'identifier',
                    {
                        'analog': 'humdrum:OPS'
                    }
                )
                identifierElement.text = opusNumber.meiValue

            # all <title>s
            theWork.subElements.extend(titleElements)

            # all <composer>s
            theWork.subElements.extend(composerElements)

            # <lyricist>
            for lyricist in lyricists:
                lyricistElement: MeiElement = theWork.appendSubElement('lyricist')
                persName = lyricistElement.appendSubElement(
                    'persName',
                    {
                        'analog': 'humdrum:LYR'
                    }
                )
                persName.text = lyricist.meiValue

            # <librettist>
            for librettist in librettists:
                librettistElement: MeiElement = theWork.appendSubElement('librettist')
                persName = librettistElement.appendSubElement(
                    'persName',
                    {
                        'analog': 'humdrum:LIB'
                    }
                )
                persName.text = librettist.meiValue

            # <funder>
            for funder in funders:
                funderElement: MeiElement = theWork.appendSubElement('funder')
                persName = funderElement.appendSubElement(
                    'name',
                    {
                        'analog': 'humdrum:OCO'
                    }
                )
                persName.text = funder.meiValue

            # <creation>
            if (creationDates
                    or creationCountries
                    or creationSettlements
                    or creationRegions
                    or creationLatLongs
                    or dedicatees):
                creationElement: MeiElement = theWork.appendSubElement('creation')

                for creationDate in creationDates:
                    dateElement: MeiElement = creationElement.appendSubElement(
                        'date',
                        {
                            'analog': 'humdrum:ODT',
                        }
                    )
                    dateElement.fillInIsodate(creationDate.value)
                    dateElement.text = creationDate.meiValue

                for creationCountry in creationCountries:
                    countryElement: MeiElement = creationElement.appendSubElement(
                        'country',
                        {
                            'analog': 'humdrum:OCY'
                        }
                    )
                    countryElement.text = creationCountry.meiValue

                for creationSettlement in creationSettlements:
                    settlementElement: MeiElement = creationElement.appendSubElement(
                        'settlement',
                        {
                            'analog': 'humdrum:OPC'
                        }
                    )
                    settlementElement.text = creationSettlement.meiValue

                for creationRegion in creationRegions:
                    regionElement: MeiElement = creationElement.appendSubElement(
                        'geogName',
                        {
                            'analog': 'humdrum:ARE'
                        }
                    )
                    regionElement.text = creationRegion.meiValue

                for creationLatLong in creationLatLongs:
                    regionElement = creationElement.appendSubElement(
                        'geogName',
                        {
                            'type': 'coordinates',
                            'analog': 'humdrum:ARL'
                        }
                    )
                    regionElement.text = creationLatLong.meiValue

                for dedicatee in dedicatees:
                    contributorElement = creationElement.appendSubElement(
                        'dedicatee',
                        {
                            'analog': 'humdrum:ODE'
                        }
                    )
                    contributorElement.text = dedicatee.meiValue

            # <history>
            if histories:
                allTheSameLanguage: bool
                theLanguage: str | None
                theLanguage, allTheSameLanguage = MeiMetadata.getTextListLanguage(histories)
                historyElement: MeiElement = theWork.appendSubElement('history')
                attrib: dict[str, str] = {}
                if allTheSameLanguage and theLanguage:
                    attrib['xml:lang'] = theLanguage
                lgElement: MeiElement = historyElement.appendSubElement('lg', attrib)

                for history in histories:
                    if t.TYPE_CHECKING:
                        assert isinstance(history.value, m21.metadata.Text)
                    attrib = {'type': 'humdrum:HAO'}
                    if history.value.language and not allTheSameLanguage:
                        attrib['xml:lang'] = history.value.language.lower()
                    # <l> can't take @analog, so use @type (says Perry)
                    lElement: MeiElement = lgElement.appendSubElement('l', attrib)
                    lElement.text = history.meiValue

            # <langUsage>
            if languages:
                langUsageElement: MeiElement = theWork.appendSubElement('langUsage')
                for lang in languages:
                    languageElement: MeiElement = langUsageElement.appendSubElement(
                        'language',
                        {
                            'analog': 'humdrum:TXO'
                        }
                    )
                    languageElement.text = lang.meiValue

            # TODO: <perfMedium><perfResList>
#             if instrumentLists:
#                 perfMediumElement: MeiElement = theWork.appendSubElement('perfMedium')
#                 if len(instrumentLists) == 1:
#                     perfResListElement: MeiElement = perfMediumElement.appendSubElement(
#                         'perfResList'
#                     )
#                     for instrument in instrumentLists[0].split(' '):
#                         perfResElement: MeiElement = perfResListElement.appendSubElement(
#                             'perfRes'
#                         )
#                         perfResElement.text = oh boy, what about counts
#                     outerPerfResListElement: MeiElement = perfMediumElement.appendSubElement(
#                         'perfResList'
#                     )

            # <classification>
            if forms or genres or modes or meters or styles:
                classificationElement: MeiElement = theWork.appendSubElement('classification')
                termListElement: MeiElement = classificationElement.appendSubElement('termList')
                for form in forms:
                    termElement: MeiElement = termListElement.appendSubElement(
                        'term',
                        {
                            'label': 'form',
                            'analog': 'humdrum:AFR'
                        }
                    )
                    termElement.text = form.meiValue

                for genre in genres:
                    termElement = termListElement.appendSubElement(
                        'term',
                        {
                            'label': 'genre',
                            'analog': 'humdrum:AGN'
                        }
                    )
                    termElement.text = genre.meiValue

                for mode in modes:
                    termElement = termListElement.appendSubElement(
                        'term',
                        {
                            'label': 'mode',
                            'analog': 'humdrum:AMD'
                        }
                    )
                    termElement.text = mode.meiValue

                for meter in meters:
                    termElement = termListElement.appendSubElement(
                        'term',
                        {
                            'label': 'meter',
                            'analog': 'humdrum:AMT'
                        }
                    )
                    termElement.text = meter.meiValue

                for style in styles:
                    termElement = termListElement.appendSubElement(
                        'term',
                        {
                            'label': 'style',
                            'analog': 'humdrum:AST'
                        }
                    )
                    termElement.text = style.meiValue

            # <expressionList>
            expressionListElement: MeiElement | None = None
            if firstPerformanceDates:
                expressionListElement = theWork.appendSubElement('expressionList')

                expressionElement: MeiElement = expressionListElement.appendSubElement(
                    'expression'
                )
                titleElement = expressionElement.appendSubElement('title')
                titleElement.text = "First performance"
                creationElement = expressionElement.appendSubElement('creation')
                for firstPerformanceDate in firstPerformanceDates:
                    dateElement = creationElement.appendSubElement(
                        'date',
                        {
                            'type': 'firstPerformance',
                            'analog': 'humdrum:MPD'
                        }
                    )
                    dateElement.fillInIsodate(firstPerformanceDate.value)
                    dateElement.text = firstPerformanceDate.meiValue

            if performanceDates:
                if expressionListElement is None:
                    expressionListElement = theWork.appendSubElement('expressionList')

                for i, performanceDate in enumerate(performanceDates):
                    expressionElement = expressionListElement.appendSubElement(
                        'expression'
                    )
                    titleElement = expressionElement.appendSubElement('title')
                    titleElement.text = 'Performance'
                    creationElement = expressionElement.appendSubElement('creation')

                    dateElement = creationElement.appendSubElement(
                        'date',
                        {
                            'type': 'performance',
                            'analog': 'humdrum:MDT'
                        }
                    )
                    dateElement.fillInIsodate(performanceDate.value)
                    dateElement.text = performanceDate.meiValue

                    if i < len(performanceLocations):
                        performanceLocation = performanceLocations[i]
                        geogNameElement = creationElement.appendSubElement(
                            'geogName',
                            {
                                'role': 'performanceLocation',
                                'analog': 'humdrum:MLC'
                            }
                        )
                        geogNameElement.text = performanceLocation.meiValue

            # <relationList> (relations to the other works)
            if parentWorkXmlId or groupWorkXmlId or associatedWorkXmlId or collectionWorkXmlId:
                relationList = theWork.appendSubElement('relationList')
                if parentWorkXmlId:
                    relationList.appendSubElement(
                        'relation',
                        {
                            'rel': 'isPartOf',
                            'type': 'isChildOfParent',
                            'target': f'#{parentWorkXmlId}'
                        }
                    )
                if groupWorkXmlId:
                    relationList.appendSubElement(
                        'relation',
                        {
                            'rel': 'isPartOf',
                            'type': 'isMemberOfGroup',
                            'target': f'#{groupWorkXmlId}'
                        }
                    )
                if associatedWorkXmlId:
                    relationList.appendSubElement(
                        'relation',
                        {
                            'rel': 'isVersionOf',
                            'type': 'isAssociatedWith',
                            'target': f'#{associatedWorkXmlId}'
                        }
                    )
                if collectionWorkXmlId:
                    relationList.appendSubElement(
                        'relation',
                        {
                            'rel': 'isPartOf',
                            'type': 'isMemberOfCollection',
                            'target': f'#{collectionWorkXmlId}'
                        }
                    )

            extMeta: MeiElement | None = self.makeExtMetaElementForWork()
            if extMeta is not None:
                theWork.subElements.append(extMeta)

        return workList

    def makeManifestationListElement(self) -> MeiElement | None:
        return None

    def makeExtMetaElementForMeiHead(self) -> MeiElement | None:
        return None

    def makeExtMetaElementForWork(self) -> MeiElement | None:
        if not self.madsCollection:
            return None

        extMeta: MeiElement = MeiElement('extMeta')
        extMeta.subElements.append(self.madsCollection)
        self.madsCollection = None

        return extMeta



# ------------------------------------------------------------------------------
# Name:          meistaff.py
# Purpose:       MeiStaff is an object that represents a single music21
#                Part/PartStaff's contributions to a particular MeiMeasure.
#
# Authors:       Greg Chapman <gregc@mac.com>
#
# Copyright:     (c) 2023 Greg Chapman
# License:       MIT, see LICENSE
# ------------------------------------------------------------------------------
import sys
from xml.etree.ElementTree import TreeBuilder
# import typing as t

import music21 as m21
# from music21.common import opFrac

# from converter21.mei import MeiExportError
# from converter21.mei import MeiInternalError
# from converter21.shared import M21Utilities
from converter21.mei import M21ObjectConvert
from converter21.mei import MeiLayer

# For debug or unit test print, a simple way to get a string which is the current function name
# with a colon appended.
# for current func name, specify 0 or no argument.
# for name of caller of current func, specify 1.
# for name of caller of caller of current func, specify 2. etc.
# pylint: disable=protected-access
funcName = lambda n=0: sys._getframe(n + 1).f_code.co_name + ':'  # pragma no cover
# pylint: enable=protected-access

class MeiStaff:
    Debug: bool = False  # can be set to True for more debugging

    def __init__(
        self,
        staffNStr: str,
        m21Measure: m21.stream.Measure,
        m21Part: m21.stream.Part,
        spannerBundle: m21.spanner.SpannerBundle,
        scoreMeterStream: m21.stream.Stream[m21.meter.TimeSignature]
    ) -> None:
        self.staffNStr: str = staffNStr
        self.m21Measure = m21Measure
        self.m21Part = m21Part
        self.spannerBundle = spannerBundle
        self.scoreMeterStream = scoreMeterStream
        self.nextFreeVoiceNumber = 1
        self.layers: list[MeiLayer] = []
        self.theOneLayerIsTheMeasureItself = False

        voices: list[m21.stream.Voice | m21.stream.Measure]
        if m21Measure.voices:
            voices = list(m21Measure.voices)
        else:
            voices = [m21Measure]
            self.theOneLayerIsTheMeasureItself = True

        for voice in voices:
            self.layers.append(MeiLayer(voice, self, spannerBundle, scoreMeterStream))

    def makeRootElement(self, tb: TreeBuilder):
        self.nextFreeVoiceNumber = 1
        if not self.theOneLayerIsTheMeasureItself:
            # Process any clef/timesig/keysig at offset 0 in enclosing measure (but
            # if the m21Measure itself is the first measure in the m21Part, skip
            # the very first clef, timesig and keysig, since those initial ones are
            # handled in the original <scoredef>).
            # We assume that any clef/timesig/keysig at non-zero measure offset will
            # be sitting alongside the notes (e.g. in a Voice), and can just be emitted
            # like a note, without this <staffdef> wrapper.
            firstSeen: list[str] = []
            staffDefEmitted: bool = False
            for el in self.m21Measure:
                if el.offset == 0:
                    if isinstance(
                        el,
                        (m21.clef.Clef, m21.meter.TimeSignature, m21.key.KeySignature)
                    ):
                        if self.m21Measure.offset == 0:
                            # very first measure in part:
                            # skip the first clef, the first timesig, and first keysig
                            if isinstance(el, m21.clef.Clef):
                                if 'clef' not in firstSeen:
                                    firstSeen.append('clef')
                                    continue
                            if isinstance(el, m21.meter.TimeSignature):
                                if 'timesig' not in firstSeen:
                                    firstSeen.append('timesig')
                                    continue
                            if isinstance(el, m21.key.KeySignature):
                                if 'keysig' not in firstSeen:
                                    firstSeen.append('keysig')
                                    continue

                        if not staffDefEmitted:
                            tb.start('staffDef', {'n': self.staffNStr})
                            staffDefEmitted = True
                        M21ObjectConvert.convertM21ObjectToMei(el, tb)
            if staffDefEmitted:
                tb.end('staffDef')

        tb.start('staff', {'n': self.staffNStr})
        for layer in self.layers:
            layer.makeRootElement(tb)
        # 888 process any final barline in the Measure (MeiStaff)
        tb.end('staff')



    def makePostStavesElements(self, tb: TreeBuilder):
        for layer in self.layers:
            layer.makePostStavesElements(tb)

        # The top-level measure (if not treated as the one layer) might have some
        # post-staves elements, too (e.g. Dynamic, TextExpression, TempoIndication).
        # This is a small subset of what can be emitted in layer.makePostStavesElements.
        # Also, there may be SpannerAnchors that exist at the Measure level.
        if not self.theOneLayerIsTheMeasureItself:
            for obj in self.m21Measure:
                if M21ObjectConvert.streamElementBelongsInPostStaves(obj):
                    M21ObjectConvert.convertPostStaveStreamElement(
                        obj,
                        self.staffNStr,
                        self.m21Part,
                        self.m21Measure,
                        self.scoreMeterStream,
                        tb
                    )
                if isinstance(obj, m21.spanner.SpannerAnchor):
                    for spanner in obj.getSpannerSites():
                        if hasattr(spanner, 'mei_trill_already_handled'):
                            continue
                        if spanner.isFirst(obj):
                            # print(f'spanner seen: {spanner.classes[0]}', file=sys.stderr)
                            M21ObjectConvert.postStavesSpannerToMei(
                                spanner,
                                self.staffNStr,
                                self.m21Part,
                                self.m21Measure,
                                self.scoreMeterStream,
                                tb
                            )

        # lastly, any fermata on the right barline is a post-staves element.
        if self.m21Measure.rightBarline is not None:
            if self.m21Measure.rightBarline.pause is not None:
                M21ObjectConvert.fermataToMei(
                    self.m21Measure.rightBarline,
                    self.m21Measure.rightBarline.pause,
                    self.staffNStr,
                    self.m21Part,
                    self.m21Measure,
                    self.scoreMeterStream,
                    tb
                )




# ------------------------------------------------------------------------------
# Name:          HumdrumFileBase.py
# Purpose:       Used to store Humdrum text lines from input stream
#                for further parsing.  This class analyzes the basic
#                spine structure after reading a Humdrum file.
#
# Authors:       Greg Chapman <gregc@mac.com>
#                Humdrum code derived/translated from humlib (authored by
#                       Craig Stuart Sapp <craig@ccrma.stanford.edu>)
#
# Copyright:     (c) 2021-2023 Greg Chapman
# License:       MIT, see LICENSE
# ------------------------------------------------------------------------------
import sys
import typing as t
from pathlib import Path

from music21.common import opFrac

from converter21.humdrum import HumdrumInternalError
from converter21.humdrum import HumHash
from converter21.humdrum import HumNum, HumNumIn
from converter21.humdrum import HumSignifiers
from converter21.humdrum import HumdrumToken
from converter21.humdrum import HumdrumLine

# For debug or unit test print, a simple way to get a string which is the current function name
# with a colon appended.
# for current func name, specify 0 or no argument.
# for name of caller of current func, specify 1.
# for name of caller of caller of current func, specify 2. etc.
# pylint: disable=protected-access
funcName = lambda n=0: sys._getframe(n + 1).f_code.co_name + ':'  # pragma no cover
# pylint: enable=protected-access

# simple class to represent a pair of (first, last) related tokens
class TokenPair:
    def __init__(
        self,
        first: HumdrumToken | None,
        last: HumdrumToken | None
    ) -> None:
        self._first: HumdrumToken | None = first
        self._last: HumdrumToken | None = last

    @property
    def first(self) -> HumdrumToken | None:
        return self._first

    @first.setter
    def first(self, newFirst: HumdrumToken) -> None:
        self._first = newFirst

    @property
    def last(self) -> HumdrumToken | None:
        return self._last

    @last.setter
    def last(self, newLast: HumdrumToken) -> None:
        self._last = newLast

    # the following two properties are used for sorting (sort by line index, then by field index)
    @property
    def firstLineIndex(self) -> int:
        if t.TYPE_CHECKING:
            assert isinstance(self._first, HumdrumToken)
        return self._first.lineIndex

    @property
    def firstFieldIndex(self) -> int:
        # Let mypy assume all is well... we'll crash appropriately if someone tries to
        # do this before populating _first.
        if t.TYPE_CHECKING:
            assert isinstance(self._first, HumdrumToken)
        return self._first.fieldIndex


class HumFileAnalysis:
    def __init__(self) -> None:
        self.clear()

    def clear(self) -> None:
        self.structureAnalyzed: bool = False
        self.rhythmAnalyzed: bool = False
        self.strandsAnalyzed: bool = False
        self.slursAnalyzed: bool = False
        self.phrasesAnalyzed: bool = False
        self.nullsAnalyzed: bool = False
        self.strophesAnalyzed: bool = False
        self.barlinesAnalyzed: bool = False
        self.barlinesDifferent: bool = False


class HumdrumFileBase(HumHash):
    '''
        Options bits for getTrackSequence() -> list[list[HumdrumToken]]
    '''
    OPT_PRIMARY: int = 0x001
    OPT_NOEMPTY: int = 0x002
    OPT_NONULL: int = 0x004
    OPT_NOINTERP: int = 0x008
    OPT_NOMANIP: int = 0x010
    OPT_NOCOMMENT: int = 0x020
    OPT_NOGLOBAL: int = 0x040
    OPT_NOREST: int = 0x080
    OPT_NOTIE: int = 0x100
    OPT_DATA: int = (OPT_NOMANIP
                            | OPT_NOCOMMENT
                            | OPT_NOGLOBAL)
    OPT_ATTACKS: int = (OPT_DATA
                            | OPT_NOREST
                            | OPT_NOTIE
                            | OPT_NONULL)

    def __init__(self, fileName: str | Path | None = None) -> None:
        super().__init__()  # initialize the HumHash fields

        '''
        // m_lines: an array representing lines from the input file.
        // The contents of lines must be deallocated when deconstructing object.
        '''
        self._lines: list[HumdrumLine] = []

        '''
        // m_filename: name of the file which was loaded.
        '''
        # self._fileName: str = None # weirdly, appears not be set or used

        '''
        // m_segementlevel: segment level (e.g., work/movement)
        '''
        self._segmentLevel: int = 0

        '''
        // m_trackstarts: list of addresses of the exclusive interpreations
        // in the file.  The first element in the list is reserved, so the
        // number of tracks (primary spines) is equal to one less than the
        // size of this list.
        '''
        self._trackStarts: list[HumdrumToken | None] = [None]

        '''
        // m_trackends: list of the addresses of the spine terminators in the
        // file. It is possible that spines can split and their subspines do not
        // merge before termination; therefore, the ends are stored in
        // a 2d array. The first dimension is the track number, and the second
        // dimension is the list of terminators.
            e.g. trackEnd2 = trackEnds[trackNum][2]
        '''
        self._trackEnds: list[list[HumdrumToken]] = [[]]

        '''
        // m_barlines: list of barlines in the data.  If the first measures is
        // a pickup measure, then the first entry will not point to the first
        // starting exclusive interpretation line rather than to a barline.
        // LATER: Maybe also add "measures" which are complete metrical cycles.
        '''
        self._barlines: list[HumdrumLine] = []

        '''
        // m_ticksperquarternote: this is the number of tick
        '''
        self._ticksPerQuarterNote: int = -1

        '''
        // m_idprefix: an XML id prefix used to avoid id collisions when
        // including multiple HumdrumFile XML in a single group.
        '''
        self._idPrefix: str = ''

        '''
        // m_strands1d: one-dimensional list of spine strands.
        '''
        self._strand1d: list[TokenPair] = []

        '''
        // m_strands2d: two-dimensional list of spine strands.
        '''
        self._strand2d: list[list[TokenPair]] = []

        '''
        // m_strophes1d: one-dimensional list of all *strophe/*Xstrophe pairs.
        '''
        self._strophes1d: list[TokenPair] = []

        '''
        // m_strophes2d: two-dimensional list of all *strophe/*Xstrophe pairs.
        '''
        self._strophes2d: list[list[TokenPair]] = []

        '''
        // m_quietParse: Set to true if error messages should not be
        // printed to the console when reading.
        '''
        self._quietParse: bool = False

        '''
        // m_parseError: Set to true if a read is successful.
        '''
        self._parseError: str = ''

        '''
        // m_displayError: Used to print error message only once.
        '''
        self._displayError: bool = False

        '''
        // m_signifiers: Used to keep track of !!!RDF records.
        '''
        self._signifiers: HumSignifiers = HumSignifiers()

        '''
            _hasInformalBreaks: True if there are '!!pagebreak:' or '!!linebreak:' breaks
                                in the file
            _hasFormalBreaks: True if there are '!LO:LB' or '!LO:PB' breaks in the file
        '''
        self._hasInformalBreaks: bool = False
        self._hasFormalBreaks: bool = False

        '''
        // m_analysis: Used to keep track of analysis states for the file.
        '''
        self._analyses: HumFileAnalysis = HumFileAnalysis()

        '''
            _spineColor: color, indexed by track and subtrack.  '' means use default color
            from iohumdrum.cpp, now computed in HumdrumFileContent.analyzeNotation()
        '''
        self._spineColor: list[list[str]] = []

        # only used by test infrastructure...
        # if we did this, then we skip the str(hf) == fileContents test
        self.fixedUpRecipOnlyToken: bool = False

        '''
            We are also a HumHash, and that was initialized via super() above.
            But we want our default HumHash prefix to be '!!', not ''
        '''
        self.prefix = '!!'

        if fileName is not None:
            self.read(fileName)

    '''
        Conversion to str.  All the lines, with '\n' between them.
    '''
    def __str__(self) -> str:
        contents: str = ''
        for i, line in enumerate(self._lines):
            if i > 0:
                contents += '\n'
            contents += line.text
        return contents

    '''
        Simple indexer, so you can do line5 = hdFile[5], without giving access to the whole array.
        Equivalent to C++: HumdrumLine& operator[](int index)
        Returns None if index is out of bounds.
    '''
    def __getitem__(self, index: int) -> HumdrumLine | None:
        if not isinstance(index, int):
            # if its a slice, out-of-range start/stop won't crash
            return self._lines[index]

        if index < 0:
            index += len(self._lines)

        if index < 0:
            return None
        if index >= len(self._lines):
            return None

        return self._lines[index]

    '''
        A generator, so clients can do: for line in hdFile.lines(), without giving them access
        to the whole array
    '''
    def lines(self):
        for lineIdx in range(0, len(self._lines)):
            yield self._lines[lineIdx]

    '''
    //////////////////////////////
    //
    // HumdrumFileBase::clear -- Reset the contents of a file to be empty.
    '''
    def clear(self) -> None:
        self._lines = []
        # self._fileName = ''
        self._trackStarts = [None]
        self._trackEnds = [[]]
        self._barlines = []
        self._segmentLevel = 0
        self._ticksPerQuarterNote = -1
        self._idPrefix = ''
        self._strand1d = []
        self._strand2d = []
        self._strophes1d = []
        self._strophes2d = []
        self._analyses.clear()

    '''
    //////////////////////////////
    //
    // HumdrumFileBase::read -- Load file contents from an input stream or file.
    '''
    def read(self, fileName: str | Path) -> bool:
        # with open(fileName, errors='backslashreplace') as f: <- this always succeeds, but...
        try:
            with open(fileName, encoding='utf-8') as f:
                s = f.read()
                return self.readString(s)

        except UnicodeDecodeError:
            with open(fileName, encoding='latin-1') as f:
                s = f.read()
                return self.readString(s)

        return False

    '''
    //////////////////////////////
    //
    // HumdrumFileBase::readString -- Read contents from a string rather than
    //    an istream or filename.
    '''
    def readString(self, contents: str) -> bool:
        if contents[-1] == '\n':
            contents = contents[0:-1]  # lose that last empty line (many editors add it)

        contentLines = contents.split('\n')
        # print(funcName(), 'len(contentLines) =', len(contentLines), file=sys.stderr)
        for contentLine in contentLines:
            line = HumdrumLine(contentLine)
            line.ownerFile = self
            self._lines.append(line)

        self.analyzeBaseFromLines()
        # print(funcName(), 'self.isValid =', self.isValid, file=sys.stderr)
        return self.isValid

    '''
    //////////////////////////////
    //
    // HumdrumFileBase::getMergedSpineInfo -- Will only simplify a two-spine
    //   merge.  Should be expanded to larger spine mergers in the future.
    //   In other words, it is best to currently merge spines in the order
    //   in which they were split, so that the original spine label can
    //   be produced.
    '''
    @staticmethod
    def getMergedSpineInfo(
        info: list[str],
        startSpine: int,
        numExtraSpines: int
    ) -> str:
        # print(funcName(), 'startSpine =', startSpine, 'numExtraSpines =', numExtraSpines,
        #       file=sys.stderr)
        # print(funcName(), 'info =', info, file=sys.stderr)

        output: str = ''
        if numExtraSpines < 1:
            # Strange if get here. Nothing to merge.
            return info[startSpine]

        if numExtraSpines == 1:
            # two-spine merge
            # compare all but trailing 'a' or 'b'
            if info[startSpine][:-1] == info[startSpine + 1][:-1]:
                # print(funcName(), 'merged spineInfo =', info[startSpine][1:-2], file=sys.stderr)
                # strip off leading '(' and the trailing ')a' or ')b'
                return info[startSpine][1:-2]
            # print(funcName(), 'merged spineInfo =', info[startSpine] + ' ' + info[startSpine+1],
            #       file=sys.stderr)
            return info[startSpine] + ' ' + info[startSpine + 1]
        '''
        // Generalized code for simplifying up to N subspines at once
        // Not fully generalized so that the subspines will always be
        // simplified if not merged in a simple way, though.
        '''
        newInfo: list[str] = [
            info[i] for i in range(startSpine, startSpine + numExtraSpines + 1)
        ]
        while len(newInfo) > 1:
            simplifiedSomething = False
            for i in range(1, len(newInfo)):
                if newInfo[i - 1][:-1] == newInfo[i][:-1]:  # compare all but trailing a or b
                    simplifiedSomething = True
                    newInfo[i - 1] = ''             # we'll remove this later
                    newInfo[i] = newInfo[i][1:-2]   # strip off leading ( and trailing )a or )b

            newInfo2: list[str] = []
            for infoStr in newInfo:
                if infoStr != '':
                    newInfo2.append(infoStr)

            for i in range(1, len(newInfo2)):
                if newInfo2[i - 1][:-1] == newInfo2[i][:-1]:  # compare all but trailing a or b
                    simplifiedSomething = True
                    newInfo2[i - 1] = ''             # we'll remove this later
                    newInfo2[i] = newInfo2[i][1:-2]  # strip off leading ( and trailing )a or ')b'

            # do it again back into newInfo
            newInfo = []
            for infoStr in newInfo2:
                if infoStr != '':
                    newInfo.append(infoStr)

            if not simplifiedSomething:
                # we've simplified as much as we can, don't loop forever
                break

        output = newInfo[0]

        # Anything left beyond newInfo[0]? Just join with ' ' as delimiter.
        for i in range(1, len(newInfo)):
            output += ' ' + newInfo[i]

        # print(funcName(), 'merged spineInfo =', output, file=sys.stderr)
        return output


    '''
    //////////////////////////////
    //
    // HumdrumFileBase::addUniqueTokens -- Used for non-null token analysis.

        Adds all tokens from source to target, unless they are already in target.
    '''
    @staticmethod
    def addUniqueTokens(
        target: list[HumdrumToken],
        source: list[HumdrumToken]
    ) -> None:
        for srcToken in source:
            found = False
            for targToken in target:
                if srcToken == targToken:
                    found = True
                    break
            if not found:
                target.append(srcToken)

    '''
    //////////////////////////////
    //
    // HumdrumFileBase::isStructureAnalyzed --
    '''
    @property
    def isStructureAnalyzed(self) -> bool:
        return self._analyses.structureAnalyzed

    '''
    //////////////////////////////
    //
    // HumdrumFileBase::isRhythmAnalyzed --
    '''
    @property
    def isRhythmAnalyzed(self) -> bool:
        return self._analyses.rhythmAnalyzed

    '''
    //////////////////////////////
    //
    // HumdrumFileBase::areStrandsAnalyzed --
    '''
    @property
    def areStrandsAnalyzed(self) -> bool:
        return self._analyses.strandsAnalyzed

    '''
    //////////////////////////////
    //
    // HumdrumFileBase::areStrophesAnalyzed --
    '''
    @property
    def areStrophesAnalyzed(self) -> bool:
        return self._analyses.strophesAnalyzed

    '''
    //////////////////////////////
    //
    // HumdrumFileBase::getXmlIdPrefix -- Return the HumdrumXML ID attribute prefix.
    // HumdrumFileBase::setXmlIdPrefix -- Set the prefix for a HumdrumXML ID
    //     atrribute.  The prefix should not start with a digit, nor have
    //     spaces in it.
    '''
    @property
    def xmlIdPrefix(self) -> str:
        return self._idPrefix

    @xmlIdPrefix.setter
    def xmlIdPrefix(self, newXMLIdPrefix: str) -> None:
        self._idPrefix = newXMLIdPrefix

    '''
    //////////////////////////////
    //
    // HumdrumFileBase::getParseError -- Return parse fail reason.
    // HumdrumFileBase::setParseError -- Set an error message from parsing
    //     input data.  The size of the message will keep track of whether
    //     or not an error was generated.  If no error message is generated
    //     when reading data, then the parsing of the data is assumed to be
    //     good.
    '''
    @property
    def parseError(self) -> str:
        return self._parseError

    @parseError.setter
    def parseError(self, err: str) -> None:
        self._parseError = err

    def setParseError(self, err: str) -> bool:
        self.parseError = err
        return err == ''

    '''
    //////////////////////////////
    //
    // HumdrumFileBase::isQuiet -- Returns true if parsing errors
    //    messages should be suppressed. By default the parsing
    //    is "noisy" and the error messages will be printed to
    //    standard error.
    // HumdrumFileBase::setQuietParsing -- Prevent error messages from
    //   being displayed when reading data.
    // @SEEALSO: setQuietParsing
    // @SEEALSO: setNoisyParsing
    '''
    @property
    def isQuiet(self) -> bool:
        return self._isQuiet

    @isQuiet.setter
    def isQuiet(self, newIsQuiet: bool) -> None:
        self._isQuiet = newIsQuiet

    '''
    //////////////////////////////
    //
    // HumdrumFileBase::isValid -- Returns true if last read was
    //     successful.
    '''
    @property
    def isValid(self) -> bool:
        if self._displayError and self.parseError != '' and not self.isQuiet:
            print(self.parseError, file=sys.stderr)
            self._displayError = False

        return self.parseError == ''

    '''
    //////////////////////////////
    //
    // HumdrumFileBase::analyzeBaseFromLines --
    '''
    def analyzeBaseFromLines(self) -> bool:
        if not self.analyzeTokens():
            return self.isValid
        if not self.analyzeLines():
            return self.isValid
        if not self.analyzeSpines():
            return self.isValid
        if not self.analyzeLinks():
            return self.isValid
        if not self.analyzeTracks():
            return self.isValid
        if not self.fixupKernRecipOnlyTokens():
            return self.isValid
        return self.isValid

    # for use by HumdrumWriter, for example, who has already created tokens and lines.
    def analyzeBase(self) -> bool:
        # this only happens in readString, so we need to do it here for exported HumdrumFiles.
        for line in self.lines():
            line.ownerFile = self
        # don't call analyzeTokens, we already have lines from the tokens

        if not self.analyzeLines():
            return self.isValid
        if not self.analyzeSpines():
            return self.isValid
        if not self.analyzeLinks():
            return self.isValid
        if not self.analyzeTracks():
            return self.isValid
        if not self.fixupKernRecipOnlyTokens():
            return self.isValid
        return self.isValid


    '''
        fixupKernRecipOnlyTokens: certain krn files (looking at you,
        rds-scores/R415_Web-w28p2-3m11-19.krn) have recip-only tokens
        in **kern spines.  These appear to be invisible duration markers
        that cover for the fact that some of the notes in this staff
        are actually annotated in other humdrum staffs.  iohumdrum.cpp
        just ignores these, but while that might work for MEI and/or
        verovio, music21 gets very confused if those "rests" are missing
        from a measure/voice.  Here, we fix this up by converting them to
        invisible rests.
    '''
    def fixupKernRecipOnlyTokens(self) -> bool:
        for line in self._lines:
            if line.isGlobal:
                continue

            fixedSomething: bool = False
            for token in line.tokens():
                if token.isKern and token.isRecipOnly:
                    token.text += 'ryy'
                    fixedSomething = True

            if fixedSomething:
                line.createLineFromTokens()
                self.fixedUpRecipOnlyToken = True

        return self.isValid

    '''
    //////////////////////////////
    //
    // HumdrumFileBase::analyzeTokens -- Generate token array from
    //    current contents of the lines.  If either tokens or the line
    //    is changed, then the other state becomes invalid.
    //    See createLinesFromTokens for regeneration of lines from tokens.
    '''
    def analyzeTokens(self) -> bool:
        for line in self._lines:
            line.createTokensFromLine()
        return self.isValid

    '''
    //////////////////////////////
    //
    // HumdrumFileBase::analyzeLines -- Store a line's index number in the
    //    HumdrumFile within the HumdrumLine object at that index.
    //    Returns false if there was an error.
    '''
    def analyzeLines(self) -> bool:
        for i, line in enumerate(self._lines):
            line.lineIndex = i
        return self.isValid

    '''
    //////////////////////////////
    //
    // HumdrumFileBase::analyzeTracks -- Analyze the track structure of the
    //     data.  Returns false if there was a parse error.
    '''
    def analyzeTracks(self) -> bool:
        for line in self._lines:
            success = line.analyzeTracks()
            if not success:
                return False

        return self.isValid

    '''
    //////////////////////////////
    //
    // HumdrumFileBase::analyzeLinks -- Generate forward and backwards spine links
    //    for each token.
    //
    '''
    def analyzeLinks(self) -> bool:
        nextLine: HumdrumLine | None = None
        prevLine: HumdrumLine | None = None
        for line in self._lines:
            if not line.hasSpines:
                continue
            prevLine = nextLine
            nextLine = line
            if prevLine is not None:
                if not self.stitchLinesTogether(prevLine, nextLine):
                    return self.isValid

        return self.isValid

    '''
        analyzeLinksForLines is for use after line insertion.  Make sure
            the range starts and ends on pre-existing (non-inserted) lines,
            and contains all inserted lines.
    '''
    def analyzeLinksForLineRange(self, rangeStart: int, rangeStop: int) -> bool:
        prevLine: HumdrumLine | None = None
        nextLine: HumdrumLine | None = None

        for lineIdx in range(rangeStart, rangeStop):
            line = self[lineIdx]
            if not line:
                continue
            if not line.hasSpines:
                continue
            prevLine = nextLine
            nextLine = line
            if prevLine is not None:
                if not self.stitchLinesTogether(prevLine, nextLine):
                    return self.isValid

        return self.isValid

    '''
    //////////////////////////////
    //
    // HumdrumFileBase::stitchLinesTogether -- Make forward/backward links for
    //    tokens on each line.
    '''
    def stitchLinesTogether(self, prevLine: HumdrumLine, nextLine: HumdrumLine) -> bool:
        # first handle simple cases where the spine assignments are one-to-one:
        if not prevLine.isInterpretation and not nextLine.isInterpretation:
            if prevLine.tokenCount != nextLine.tokenCount:
                return self.setParseError(
                    f'Error lines {prevLine.lineNumber} and {nextLine.lineNumber} '
                    + f'not same length.\nLine {prevLine.lineNumber}: {prevLine.text}\n'
                    + f'Line {nextLine.lineNumber}: {nextLine.text}\n'
                )

            for i, prevTok in enumerate(prevLine.tokens()):
                if nextLine[i] is not None:
                    prevTok.makeForwardLink(nextLine[i])
                else:
                    print('Strange error 1', file=sys.stderr)

            return True

        # Complicated case: spine manipulators have wreaked havoc
        nextTokenIdx: int = 0
        skipOneToken: bool = False
        mergeCount: int = 0
        for i, prevTok in enumerate(prevLine.tokens()):
            if skipOneToken:
                # print('skip one token', file=sys.stderr)
                skipOneToken = False
                continue

            if mergeCount != 0 and prevTok.isMergeInterpretation:
                # Still processing adjacent merge interpretations ('*v').
                # Connect this new adjacent *v spine manipulator to the
                # current next token.
                # print(funcName(), 'found another adjacent merge, i = {}'.format(i))
                if nextLine[nextTokenIdx] is not None:
                    prevTok.makeForwardLink(nextLine[nextTokenIdx])
                else:
                    print('Strange error 1.5', file=sys.stderr)

                mergeCount += 1

                if i != prevLine.tokenCount - 1:
                    continue

                # we're on the last token of the line, so fall through to finish up the merge group
                # print(funcName(), 'falling through because that last merge was at end of line')

            if mergeCount != 0 and (
                    (not prevTok.isMergeInterpretation) or (i == prevLine.tokenCount - 1)):
                # stop processing adjacent merge fields ('*v')
                # print(funcName(), 'last merge in adjacent group noticed at i = {}'.format(i))
                if mergeCount == 1:
                    # print(funcName(), 'bad *v count', file=sys.stderr)
                    return self.setParseError(
                        'Error: single spine merge indicator \'*v\' on line: '
                        + f'{prevLine.lineNumber}\n{prevLine.text}'
                    )
                nextTokenIdx += 1
                # stop counting merges, and fall through to process this non-merge token
                # if this was a merge token (i.e. it was last on the line), don't fall through
                mergeCount = 0
                if prevTok.isMergeInterpretation:
                    continue

            if not prevTok.isManipulator:
                if nextLine[nextTokenIdx] is not None:
                    prevTok.makeForwardLink(nextLine[nextTokenIdx])
                    nextTokenIdx += 1
                else:
                    print('Strange error 2', file=sys.stderr)

            elif prevTok.isSplitInterpretation:
                # Connect the previous token to the next two tokens.
                if nextLine[nextTokenIdx] is not None:
                    prevTok.makeForwardLink(nextLine[nextTokenIdx])
                    nextTokenIdx += 1
                else:
                    print('Strange error 3', file=sys.stderr)

                if nextLine[nextTokenIdx] is not None:
                    prevTok.makeForwardLink(nextLine[nextTokenIdx])
                    nextTokenIdx += 1
                else:
                    print('Strange error 4', file=sys.stderr)

            elif prevTok.isMergeInterpretation:
                # Start processing adjacent merge interpretations.
                # We will connect multiple previous tokens which are adjacent *v
                # spine manipulators to the current next token.
                # Here we do the first one, and the rest are done at the top of
                # the prevTok loop.
                # print(funcName(), 'prevLine {}: {}'.format(prevLine.lineNumber, prevLine.text),
                #       file=sys.stderr)
                # print(funcName(), 'nextLine {}: {}'.format(nextLine.lineNumber, nextLine.text),
                #       file=sys.stderr)
                # print(funcName(), 'found first adjacent merge at i={}'.format(i),
                #       file=sys.stderr)
                # print(funcName(), 'nextTokenIdx =', nextTokenIdx, file=sys.stderr)
                if nextLine[nextTokenIdx] is not None:
                    prevTok.makeForwardLink(nextLine[nextTokenIdx])
                else:
                    print('Strange error 5', file=sys.stderr)
                    raise HumdrumInternalError
                mergeCount = 1

            elif prevTok.isExchangeInterpretation:
                # swapping the order of two spines.
                prevTokPlus1: HumdrumToken | None = prevLine[i + 1]
                if prevTokPlus1 is None or not prevTokPlus1.isExchangeInterpretation:
                    return self.setParseError(
                        'Error: single spine exchange indicator \'*x\' on line: '
                        + f'{prevLine.lineNumber}\n{prevLine.text}'
                    )

                nextNext: HumdrumToken | None = nextLine[nextTokenIdx]
                if nextNext is not None:
                    prevTokPlus1.makeForwardLink(nextNext)
                else:
                    print('Strange error 6', file=sys.stderr)
                nextNextPlus1: HumdrumToken | None = nextLine[nextTokenIdx + 1]
                if nextNextPlus1 is not None:
                    prevTok.makeForwardLink(nextNextPlus1)
                else:
                    print('Strange error 7', file=sys.stderr)
                nextTokenIdx += 2
                skipOneToken = True  # we already processed prevTok[i+1], so skip one prevTok

            elif prevTok.isTerminateInterpretation:
                # No link should be made.  There may be a problem if a
                # new segment is given (this should be handled by a
                # HumdrumSet class, not HumdrumFileBase.
                pass

            elif prevTok.isAddInterpretation:
                # A new data stream is being added, the next linked token
                # should be an exclusive interpretation.
                nextTok: HumdrumToken | None = nextLine[nextTokenIdx]
                nextTokPlus1: HumdrumToken | None = nextLine[nextTokenIdx + 1]
                if nextTokPlus1 is None or not nextTokPlus1.isExclusiveInterpretation:
                    return self.setParseError(
                        'Error: expecting exclusive interpretation on line '
                        + f'{nextLine.lineNumber} at token {i}, but got {nextTokPlus1}'
                    )
                if nextTok is not None:
                    prevTok.makeForwardLink(nextTok)
                    nextTokenIdx += 1
                else:
                    print('Strange error 8', file=sys.stderr)
                nextTokenIdx += 1

            elif prevTok.isExclusiveInterpretation:
                if nextLine[nextTokenIdx] is not None:
                    prevTok.makeForwardLink(nextLine[nextTokenIdx])
                    nextTokenIdx += 1
                else:
                    print('Strange error 9', file=sys.stderr)

            else:
                return self.setParseError('Error: should not get here')

        if nextTokenIdx != nextLine.tokenCount:
            return self.setParseError(
                f'''Error: cannot stitch lines together due to alignment problem.
Line {prevLine.lineNumber}: {prevLine.text}
Line {nextLine.lineNumber}: {nextLine.text}
nextTokenIdx = {nextTokenIdx}, nextLine.tokenCount = {nextLine.tokenCount}'''
            )

        return self.isValid

    '''
    //////////////////////////////
    //
    // HumdrumFileBase::analyzeSpines -- Analyze the spine structure of the
    //     data.  Returns false if there was a parse error.
    '''
    def analyzeSpines(self) -> bool:
        dataType: list[str] = []
        sinfo: list[str] = []
        lastSpine: list[list[HumdrumToken]] = []

        self._trackStarts = [None]
        self._trackEnds = [[]]

        seenFirstExInterp = False
        for i, line in enumerate(self._lines):
            if not line.hasSpines and line[0] is not None:
                line[0].fieldIndex = 0  # the only token on the line has fieldIndex 0
                continue

            if not seenFirstExInterp and not line.isExclusiveInterpretation:
                return self.setParseError(
                    f'Error on line: {i+1}:\n'
                    + 'Data found before exclusive interpretation\n'
                    + f'LINE: {line.text}'
                )

            if not seenFirstExInterp and line.isExclusiveInterpretation:
                # first line of data in file
                seenFirstExInterp = True
                dataType = []
                sinfo = []
                for j, token in enumerate(line.tokens()):
                    dataType.append(token.text)
                    self.addToTrackStarts(token)
                    sinfo.append(str(j + 1))
                    token.spineInfo = str(j + 1)
                    token.fieldIndex = j
                    lastSpine.append(token)
                continue

            if len(dataType) != line.tokenCount:
                err = (
                    f'Error on line {line.lineNumber}:\n'
                    + f'Expected {len(dataType)} fields, but found {line.tokenCount}\n'
                    + f'Line is: {line.text}'
                )
                if i > 0:
                    err += f'\nPrevious line is {self._lines[i-1].text}'
                return self.setParseError(err)

            for j, token in enumerate(line.tokens()):
                token.spineInfo = sinfo[j]
                token.fieldIndex = j

            if line.isManipulator:
                success, dataType, sinfo = self.adjustSpines(line, dataType, sinfo)
                if not success:
                    return self.isValid

        return self.isValid

    '''
        HumdrumFileBase::getSpineCount --
    '''
    @property
    def spineCount(self) -> int:
        return self.maxTrack

    '''
    //////////////////////////////
    //
    // HumdrumFileBase::addToTrackStarts -- A starting exclusive interpretation was
    //    found, so store in the list of track starts.  The first index position
    //    in trackstarts is reserve for non-spine usage.
    '''
    def addToTrackStarts(self, token: HumdrumToken | None) -> None:
        if token is None:
            self._trackStarts.append(None)
            self._trackEnds.append([])
        elif len(self._trackStarts) > 1 and self._trackStarts[-1] is None:
            self._trackStarts[-1] = token
        else:
            self._trackStarts.append(token)
            self._trackEnds.append([])

    '''
    //////////////////////////////
    //
    // HumdrumFileBase::adjustSpines -- adjust dataType and spineInfo values based
    //   on manipulators found in the data.
    '''
    def adjustSpines(
        self,
        line: HumdrumLine,
        dataType: list[str],
        spineInfo: list[str]
    ) -> tuple[bool, list[str], list[str]]:
        # returns success, as well as the newType array, and the newInfo array
        # if there is an error, it returns success=False, and both arrays empty
        newType: list[str] = []
        newInfo: list[str] = []
        mergeCount: int = 0
        skipOneToken: bool = False

        # print(funcName(), 'line {}: {}'.format(line.lineNumber, line.text), file=sys.stderr)
        # print(funcName(), 'starting dataType =', dataType, file=sys.stderr)
        # print(funcName(), 'starting spineInfo =', spineInfo, file=sys.stderr)

        for i, token in enumerate(line.tokens()):
            if skipOneToken:
                # print(funcName(), 'skipping one token', file=sys.stderr)
                skipOneToken = False
                continue

            if mergeCount > 0 and token.isMergeInterpretation:
                mergeCount += 1
                # print(funcName(), 'found merge token #{} on this line at i = {}'.format(
                #       mergeCount, i), file=sys.stderr)

                if i != line.tokenCount - 1:
                    continue

                # we're on the last token of the line, so fall through to finish up the merge group

            if mergeCount > 0 and (
                    (not token.isMergeInterpretation) or (i == line.tokenCount - 1)):
                # print(funcName(), 'end of merge token group on this line, count =', mergeCount,
                #       file=sys.stderr)
                if mergeCount == 1:
                    # print(funcName(), 'bad *v count', file=sys.stderr)
                    self.setParseError(
                        'Error: single spine merge indicator \'*v\' on line: '
                        + f'{line.lineNumber}\n{line.text}')
                    return (False, [], [])

                # print('spineInfo = ', spineInfo, 'i-mergeCount =', i-mergeCount,
                #       'mergeCount-1 =', mergeCount-1, file=sys.stderr)
                startSpine: int = i - mergeCount
                if token.isMergeInterpretation:
                    # we stopped at the end, so i is one less than usual
                    startSpine += 1
                newInfo.append(self.getMergedSpineInfo(spineInfo, startSpine, mergeCount - 1))
                # print('mergedSpineInfo =', newInfo[-1], file=sys.stderr)
                newType.append(dataType[startSpine])
                # print('new dataType =', newType[-1], file=sys.stderr)
                # stop counting merges, and fall through to process this non-merge token
                # if this was a merge token (i.e. it was last on the line), don't fall through
                mergeCount = 0
                if token.isMergeInterpretation:
                    continue

            if token.isSplitInterpretation:
                newType.append(dataType[i])
                newType.append(dataType[i])
                newInfo.append('(' + spineInfo[i] + ')a')
                newInfo.append('(' + spineInfo[i] + ')b')
            elif token.isMergeInterpretation:
                mergeCount = 1
                # print(funcName(), 'line with merge at i = {}: {}'.format(i, line.text),
                #       file=sys.stderr)
            elif token.isAddInterpretation:
                newType.append(dataType[i])
                newType.append('')
                newInfo.append(spineInfo[i])
                self.addToTrackStarts(None)
                newInfo.append(str(self.maxTrack))
            elif token.isExchangeInterpretation:
                if i >= line.tokenCount - 1:
                    self.setParseError("Error: *x is all alone at end of line")
                    return (False, [], [])
                nextTok: HumdrumToken | None = line[i + 1]
                if t.TYPE_CHECKING:
                    # we know nextTok is not None because of i range check above
                    assert isinstance(nextTok, HumdrumToken)
                if not nextTok.isExchangeInterpretation:  # line[index] is index'th token
                    self.setParseError('Error: *x is all alone')
                    return (False, [], [])
                newType.append(dataType[i + 1])
                newType.append(dataType[i])
                newInfo.append(spineInfo[i + 1])
                newInfo.append(spineInfo[i])
                skipOneToken = True  # we already processed it here
            elif token.isTerminateInterpretation:
                # store pointer to terminate token in trackends
                # print(funcName(), 'trackStarts = {}, trackEnds = {}'.format(
                #       self._trackStarts, self._trackEnds), file=sys.stderr)
                self._trackEnds[len(self._trackStarts) - 1].append(token)
            elif token.isExclusiveInterpretation:
                newType.append(token.text)
                newInfo.append(spineInfo[i])
                if not (len(self._trackStarts) > 1 and self._trackStarts[-1] is None):
                    self.setParseError(
                        'Error: Exclusive interpretation with no preparation '
                        + f'on line {line.lineNumber} spine index {i}\n'
                        + f'Line: {line.text}'
                    )
                    return (False, [], [])
                if self._trackStarts[-1] is None:
                    self.addToTrackStarts(token)
            else:
                # should only be null interpretation
                newType.append(dataType[i])
                newInfo.append(spineInfo[i])

        # print(funcName(), 'ending dataType =', newType, file=sys.stderr)
        # print(funcName(), 'ending spineInfo =', newInfo, file=sys.stderr)
        return (True, newType, newInfo)

    '''
    //////////////////////////////
    //
    // HumdrumFileBase::createLinesFromTokens -- Generate Humdrum lines strings
    //   from the stored list of tokens.
    '''
    def createLinesFromTokens(self) -> None:
        for line in self._lines:
            line.createLineFromTokens()

    '''
    ////////////////////////////
    //
    // HumdrumFileBase::appendLine -- Add a line to the file's contents.  The file's
    //    spine and rhythmic structure should be recalculated after an append.
    '''
    def appendLine(
        self,
        line: HumdrumLine | str,
        asGlobalToken: bool = False,
        analyzeTokenLinks: bool = False
    ) -> None:
        if isinstance(line, str):
            line = HumdrumLine(line, asGlobalToken=asGlobalToken)
        if not isinstance(line, HumdrumLine):
            raise TypeError('appendLine must receive either line: str or line: HumdrumLine')
        self._lines.append(line)

        # if requested, re-analyze token links around the insertion
        if analyzeTokenLinks:
            startIdx: int = self.lineCount - 2  # previous last line
            endIdx: int = self.lineCount - 1    # new last line
            self.analyzeLinksForLineRange(startIdx, endIdx + 1)

    '''
    ////////////////////////////
    //
    // HumdrumFileBase::insertLine -- Add a line to the file's contents.  The file's
    //    spine and rhythmic structure should be recalculated after an append.
    '''
    def insertLine(
        self,
        index: int,
        aLine: HumdrumLine | str,
        asGlobalToken: bool = False,
        analyzeTokenLinks: bool = False
    ) -> None:
        if isinstance(aLine, str):
            aLine = HumdrumLine(aLine, asGlobalToken=asGlobalToken)
        if not isinstance(aLine, HumdrumLine):
            raise TypeError('insertLine must receive either line: str or line: HumdrumLine')
        self._lines.insert(index, aLine)

        # update line indexes for this line and the following ones
        for i, line in enumerate(self._lines):
            if i >= index:
                line.lineIndex = i

        # if requested, re-analyze token links around the insertion
        if analyzeTokenLinks:
            startIdx: int = max(0, index)
            endIdx: int = min(index + 1, self.lineCount - 1)
            self.analyzeLinksForLineRange(startIdx, endIdx + 1)

    '''
    //////////////////////////////
    //
    // HumdrumFileBase::insertNullDataLine -- Add a null data line at
    //     the given absolute quarter-note timestamp in the file.  If there
    //     is already a data line at the given timestamp, then do not create
    //     a line and instead return a pointer to the existing line.  Returns
    //     NULL if there was a problem.
    '''
    def insertNullDataLine(self, timestamp: HumNumIn) -> HumdrumLine | None:
        ts: HumNum = opFrac(timestamp)
        # for now do a linear search for the insertion point, but later
        # do something more efficient.
        beforet: HumNum | None = None
        beforei: int | None = None

        for i, line in enumerate(self.lines()):
            if not line.isData:
                continue

            current: HumNum = line.durationFromStart
            if current == ts:
                return line

            if current > ts:
                break

            beforet = current
            beforei = i

        if beforei is None:
            return None

        # This check is currently not necessary, since beforei and beforet are
        # both set at the same time, but will that always be true?
        # Also, mypy isn't smart enough to know, so this is a mypy hint.
        if beforet is None:
            return None

        beforeLine: HumdrumLine | None = self[beforei]
        if t.TYPE_CHECKING:
            # we know beforeLine is not None because we found it above
            # in self.lines() which never returns None.
            assert isinstance(beforeLine, HumdrumLine)

        newLine: HumdrumLine = HumdrumLine()

        # copyStructure will add null tokens automatically
        newLine.copyStructure(beforeLine, '.')

        self.insertLine(beforei + 1, newLine)

        # Set the timestamp information for inserted line:
        delta: HumNum = opFrac(ts - beforet)

        newLine.durationFromStart = opFrac(beforeLine.durationFromStart + delta)
        newLine.durationFromBarline = opFrac(beforeLine.durationFromBarline + delta)
        newLine.durationToBarline = opFrac(beforeLine.durationToBarline - delta)
        newLine.duration = opFrac(beforeLine.duration - delta)
        beforeLine.duration = delta

        for beforeToken, newToken in zip(beforeLine.tokens(), newLine.tokens()):
            beforeToken.insertTokenAfter(newToken)

        return newLine

    '''
    //////////////////////////////
    //
    // HumdrumFileBase::insertNullInterpretationLine -- Add a null interpretation
    //     line at the given absolute quarter-note timestamp in the file.  The line will
    //     be added after any other interpretation lines at that timestamp, but before any
    //     local comments that appear immediately before the data line(s) at that timestamp.
    //     Returns NULL if there was a problem.
    '''
    def insertNullInterpretationLine(self, timestamp: HumNumIn) -> HumdrumLine | None:
        ts: HumNum = opFrac(timestamp)
        # for now do a linear search for the insertion point, but later
        # do something more efficient.
        beforei: int | None = None

        for i, line in enumerate(self.lines()):
            if not line.isData:
                continue

            current: HumNum = line.durationFromStart
            if current == ts:
                beforei = i
                break

            if current > ts:
                break

            beforei = i

        if beforei is None:
            return None

        targetLine: HumdrumLine | None = self.getLineForInterpretationInsertion(beforei)
        if t.TYPE_CHECKING:
            # we know that targetLine is not None, because beforei is a valid line.
            assert isinstance(targetLine, HumdrumLine)
        newLine: HumdrumLine = HumdrumLine()
        # copyStructure will add null tokens automatically
        newLine.copyStructure(targetLine, '*')

        targeti: int = targetLine.lineIndex
        self.insertLine(targeti, newLine)

        # inserted line will increment beforei by one:
        beforei += 1
        beforeLine: HumdrumLine | None = self[beforei]
        if t.TYPE_CHECKING:
            # we know beforeLine is not None because we know that beforei
            # is in range.  We found it in range, then incremented it after
            # inserting a line before it (so it's still in range).
            assert isinstance(beforeLine, HumdrumLine)

        newLine.durationFromStart = beforeLine.durationFromStart
        newLine.durationFromBarline = beforeLine.durationFromBarline
        newLine.durationToBarline = beforeLine.durationToBarline
        newLine.duration = opFrac(0)

        # Problems here if targetLine is a manipulator
        for targetToken, newToken in zip(targetLine.tokens(), newLine.tokens()):
            targetToken.insertTokenAfter(newToken)

        return newLine

    '''
    //////////////////////////////
    //
    // HumdrumFileBase::insertNullInterpretationLineAbove -- Add a null interpretation
    //     line at the given absolute quarter-note timestamp in the file.  The line will
    //     be added before any other lines at that timestamp.
    //     Returns NULL if there was a problem.
    '''
    def insertNullInterpretationLineAbove(self, timestamp: HumNumIn) -> HumdrumLine | None:
        ts: HumNum = opFrac(timestamp)
        beforei: int | None = None

        for i, line in enumerate(self.lines()):
            current: HumNum = line.durationFromStart
            if current == ts:
                beforei = i
                break
            if current > ts:
                break

            beforei = i

        if beforei is None:
            return None

        targetLine: HumdrumLine | None = self.getLineForInterpretationInsertionAbove(beforei)
        if t.TYPE_CHECKING:
            # we know that targetLine is not None, because beforei is a valid line.
            assert isinstance(targetLine, HumdrumLine)

        newLine: HumdrumLine = HumdrumLine()
        newLine.copyStructure(targetLine, '*')

        targeti: int = targetLine.lineIndex

        self.insertLine(targeti, newLine)

        beforei += 1
        beforeLine: HumdrumLine | None = self[beforei]
        if t.TYPE_CHECKING:
            # we know beforeLine is not None because we know that beforei
            # is in range.  We found it in range, then incremented it after
            # inserting a line well before it (so it's still in range).
            assert isinstance(beforeLine, HumdrumLine)

        newLine.durationFromStart = beforeLine.durationFromStart
        newLine.durationFromBarline = beforeLine.durationFromBarline
        newLine.durationToBarline = beforeLine.durationToBarline
        newLine.duration = opFrac(0)

        for targetToken, newToken in zip(targetLine.tokens(), newLine.tokens()):
            targetToken.insertTokenAfter(newToken)

        return newLine

    '''
        insertNullInterpretationLineAt(lineIndex) does exactly that, including
        hooking up the inter-line token links. The structure of the null interp
        line is determined by copying the line that is currently at lineIndex
        before the insertion.
        The newly created null interp line is returned.
    '''
    def insertNullInterpretationLineAt(self, lineIndex: int) -> HumdrumLine:
        newLine: HumdrumLine = HumdrumLine()

        if self.lineCount <= 0:
            raise HumdrumInternalError('Cannot insert a null interpretation line in an empty file')
        if lineIndex > self.lineCount:
            raise HumdrumInternalError('Cannot insert a null interpretion line beyond the EOF')

        if lineIndex < self.lineCount:
            # insert the new line
            nextLine: HumdrumLine | None = self[lineIndex]
            if t.TYPE_CHECKING:
                # we know nextLine is not None because lineIndex < self.lineCount
                assert isinstance(nextLine, HumdrumLine)

            copyLine: HumdrumLine | None = nextLine
            if not nextLine.hasSpines:
                copyLine = self[lineIndex - 1]  # problem if previous line is manipulator
            if copyLine is None or not copyLine.hasSpines:
                raise HumdrumInternalError(
                    'Cannot insert a null interpretation line between two unspined lines'
                )
            newLine.copyStructure(copyLine, '*')
            newLine.durationFromStart = nextLine.durationFromStart
            newLine.durationFromBarline = nextLine.durationFromBarline
            newLine.durationToBarline = nextLine.durationToBarline
            newLine.duration = opFrac(0)
            self.insertLine(lineIndex, newLine, analyzeTokenLinks=True)
        else:
            # append the new line
            prevLine: HumdrumLine | None = self[lineIndex - 1]
            if t.TYPE_CHECKING:
                # we know prevLine is not None because we are appending, and if there
                # is no previous line, then we are appending to an empty file, and
                # we checked that first thing (if self.lineCount <= 0).
                assert isinstance(prevLine, HumdrumLine)

            newLine.copyStructure(prevLine, '*')  # problem if prevLine is manipulator
            newLine.durationFromStart = opFrac(prevLine.durationFromStart + prevLine.duration)
            newLine.durationFromBarline = opFrac(prevLine.durationFromBarline + prevLine.duration)
            newLine.durationToBarline = opFrac(prevLine.durationToBarline + prevLine.duration)
            newLine.duration = opFrac(0)
            self.appendLine(newLine, analyzeTokenLinks=True)

        return newLine

    '''
    //////////////////////////////
    //
    // HumdrumFileBase::getLineForInterpretationInsertion --  Search backwards
    //    in the file for the first local comment immediately before a data line
    //    index given as input.  If there are no local comments, then return the
    //    data line.  If there are local comment lines immediately before the data
    //    line, then keep searching for the first local comment.  Non-spined lines
    //    (global or empty lines) are ignored.  This function is used to insert
    //    an empty interpretation before a data line at a specific data line.
    '''
    def getLineForInterpretationInsertion(self, index: int) -> HumdrumLine | None:
        if index not in range(0, self.lineCount):
            return None

        current: int = index - 1
        previous: int = index
        while current > 0:
            currentLine: HumdrumLine | None = self[current]
            if t.TYPE_CHECKING:
                # we know currentLine is not None because current is in range
                assert isinstance(currentLine, HumdrumLine)

            if not currentLine.hasSpines:
                current -= 1
                continue
            if currentLine.isLocalComment:
                previous = current
                current -= 1
                continue

            return self[previous]

        return self[index]

    '''
    //////////////////////////////
    //
    // HumdrumFileBase::getLineForInterpretationInsertionAbove --  Search backwards
    //    in the file for the first line at the same timestamp as the starting line.
    '''
    def getLineForInterpretationInsertionAbove(self, index: int) -> HumdrumLine | None:
        if index not in range(0, self.lineCount):
            return None

        timestamp: HumNum = self._lines[index].durationFromStart
        current: int = index - 1
        previous: int = index
        while current > 0:
            currentLine: HumdrumLine | None = self[current]
            if t.TYPE_CHECKING:
                # we know currentLine is not None because current is in range
                assert isinstance(currentLine, HumdrumLine)

            if not currentLine.hasSpines:
                current -= 1
                continue
            teststamp: HumNum = currentLine.durationFromStart
            if teststamp == timestamp:
                previous = current
                current -= 1
                continue

            return self[previous]

        return self[index]

    '''
    ////////////////////////////
    //
    // HumdrumFileBase::getLineCount -- Returns the number of lines.
    '''
    @property
    def lineCount(self) -> int:
        return len(self._lines)

    '''
    //////////////////////////////
    //
    // HumdrumFileBase::getReferenceRecords --
    '''
    def referenceRecords(self) -> list[HumdrumLine]:
        refLines = []
        for line in self._lines:
            if line.isReference:
                refLines.append(line)

        return refLines

    def getReferenceValueForKey(self, key: str) -> str:
        for line in self.referenceRecords():
            if line.referenceKey == key:
                return line.referenceValue
        return ''

    '''
    //////////////////////////////
    //
    // HumdrumFileBase::getGlobalReferenceRecords --
    '''
    def globalReferenceRecords(self) -> list[HumdrumLine]:
        refLines = []
        for line in self._lines:
            if line.isGlobalReference:
                refLines.append(line)

        return refLines

    def getGlobalReferenceValueForKey(self, key: str) -> str:
        for line in self.globalReferenceRecords():
            if line.referenceKey == key:
                return line.referenceValue
        return ''

    '''
    //////////////////////////////
    //
    // HumdrumFileBase::getUniversalReferenceRecords --
    '''
    def universalReferenceRecords(self) -> list[HumdrumLine]:
        refLines = []
        for line in self._lines:
            if line.isUniversalReference:
                refLines.append(line)

        return refLines

    def getUniversalReferenceValueForKey(self, key: str) -> str:
        for line in self.universalReferenceRecords():
            if line.referenceKey == key:
                return line.referenceValue
        return ''

    '''
    //////////////////////////////
    //
    // HumdrumFileBase::token -- Return the token at the given line/field index.
    //

        Clients can just do:
            tok = hdFile[lineindex][tokenindex]
    '''

    '''
    //
    // Special case that returns a subtoken string:
    //

        Clients can just do:
            subTok = hdFile[lineindex][tokenindex].subTokens[subTokenIndex]
    '''

    '''
    //////////////////////////////
    //
    // HumdrumFileBase::getMaxTrack -- Returns the number of primary
    //     spines in the data.
    '''
    @property
    def maxTrack(self) -> int:
        return len(self._trackStarts) - 1

    '''
    //////////////////////////////
    //
    // HumdrumFileBase::getExinterpCount -- return the number of spines in
    //    the file that are of the given exinterp type.  The input string
    //    may optionally include the ** exinterp prefix.
    '''
    def getExInterpCount(self, exInterp: str) -> int:
        return len(self.spineStartListOfType(exInterp))

    '''
    //////////////////////////////
    //
    // HumdrumFileBase::getSpineStopList -- Return a list of the ending
    //     points of spine strands.
    '''
    @property
    def spineStopList(self) -> list[HumdrumToken]:
        return [trackEnd for trackEndList in self._trackEnds for trackEnd in trackEndList]

    '''
    //////////////////////////////
    //
    // HumdrumFileBase::getSpineStartList -- Return a list of the exclustive
    //     interpretations starting spines in the data.  The single parameter
    //     version of the fuction returns all starting exclusive interpretations.
    //     The two-parameter version will result all exclusive interpretations
    //     of a given datatype, and the three-parameter version where the third
    //     parameter is a vector of string, will selectively include all starting
    //     tokens which match one of the data types in the input list.  The
    //     trackstarts class variable contains an empty slot at index 0;
    //     this is removed in the return vector.
    '''
    @property
    def spineStartList(self) -> list[HumdrumToken | None]:
        return self._trackStarts[1:]

    def spineStartListOfType(self, exInterps: str | list[str]) -> list[HumdrumToken]:
        output: list[HumdrumToken] = []

        if isinstance(exInterps, str):
            exInterps = [exInterps]  # convert to list[str]

        newExInterps = [''] * len(exInterps)
        for i, exInterp in enumerate(exInterps):
            if exInterp.startswith('**'):
                newExInterps[i] = exInterp
            else:
                newExInterps[i] = '**' + exInterp

        for i, trackStart in enumerate(self._trackStarts):
            if i == 0:
                continue  # skip the first entry (there is no track 0)
            if trackStart is None:
                continue

            for exInterp in newExInterps:
                if exInterp == trackStart.text:
                    output.append(trackStart)
                    break

        return output

    def kernSpineStartList(self) -> list[HumdrumToken]:
        return self.spineStartListOfType('**kern')

    '''
    //////////////////////////////
    //
    // getPrimaryspineSequence -- Return a list of the HumdrumTokens in a spine,
    //    but not any secondary spine content if the spine splits.

        Note that spine numbers start at 0, and track numbers start at 1
    '''
    def getPrimarySpineSequence(self, spine: int, options: int) -> list[HumdrumToken]:
        track = spine + 1
        return self.getPrimaryTrackSequence(track, options)

    '''
    //////////////////////////////
    //
        getSpineSequence -- Return a list of the HumdrumTokens in a spine,
            including secondary spine content if the spine splits.

            Clients can specify spine, or startToken.  We will use spine if
            the client specifies both.
    '''
    def getSpineSequence(
        self,
        spine: int,
        startToken: HumdrumToken | None = None,
        options: int = 0
    ) -> list[list[HumdrumToken]]:
        if startToken is not None:
            track = startToken.track  # get the track number from the token
        else:
            track = spine + 1  # get the track number from the spine number
        return self.getTrackSequence(track=track, options=options)

    '''
    //////////////////////////////
    //
    // HumdrumFileBase::getPrimaryTrackSequence -- Return a list of the
    //     given primary spine tokens for a given track (indexed starting at
    //     one and going through getMaxTrack().
    '''
    def getPrimaryTrackSequence(self, track: int, options: int) -> list[HumdrumToken]:
        tempSeq: list[list[HumdrumToken]] = self.getTrackSequence(
            track=track,
            options=(options | self.OPT_PRIMARY)
        )
        return [tokenList[0] for tokenList in tempSeq]

    '''
    /////////////////////////////
    //
    // HumdrumFileBase::getTrackSequence -- Extract a sequence of tokens
    //    for the given spine.  All subspine tokens will be included.
    //    See getPrimaryTrackSequence() if you only want the first subspine for
    //    a track on all lines.
    //
    // The following options are used for the getPrimaryTrackTokens:
    // * OPT_PRIMARY    => only extract primary subspine/subtrack.
    // * OPT_NOEMPTY    => don't include null tokens in extracted list if all
    //                        extracted subspines contains null tokens.
    //                        Includes null interpretations and comments as well.
    // * OPT_NONULL     => don't include any null tokens in extracted list.
    // * OPT_NOINTERP   => don't include interpretation tokens.
    // * OPT_NOMANIP    => don't include spine manipulators (*^, *v, *x, *+,
    //                        but still keep ** and *0).
    // * OPT_NOCOMMENT  => don't include comment tokens.
    // * OPT_NOGLOBAL   => don't include global records (global comments, reference
    //                        records, and empty lines). In other words, only return
    //                        a list of tokens from lines which hasSpines() it true.
    // * OPT_NOREST     => don't include **kern rests.
    // * OPT_NOTIE      => don't include **kern secondary tied notes.
    // Compound options:
    // * OPT_DATA      (OPT_NOMANIP | OPT_NOCOMMENT | OPT_NOGLOBAL)
    //     Only data tokens (including barlines)
    // * OPT_ATTACKS   (OPT_DATA | OPT_NOREST | OPT_NOTIE | OPT_NONULL)
    //     Only note-attack tokens (when etracting **kern data)
    '''
    def getTrackSequence(
        self,
        track: int | None = None,
        startToken: HumdrumToken | None = None,
        options: int = 0
    ) -> list[list[HumdrumToken]]:
        output: list[list[HumdrumToken]] = []

        if startToken is not None:
            track = startToken.track  # get the track number from the token

        optionPrimary: bool = (options & self.OPT_PRIMARY) == self.OPT_PRIMARY
        optionNoNull: bool = (options & self.OPT_NONULL) == self.OPT_NONULL
        optionNoEmpty: bool = (options & self.OPT_NOEMPTY) == self.OPT_NOEMPTY
        optionNoInterp: bool = (options & self.OPT_NOINTERP) == self.OPT_NOINTERP
        optionNoManip: bool = (options & self.OPT_NOMANIP) == self.OPT_NOMANIP
        optionNoComment: bool = (options & self.OPT_NOCOMMENT) == self.OPT_NOCOMMENT
        optionNoGlobal: bool = (options & self.OPT_NOGLOBAL) == self.OPT_NOGLOBAL
        optionNoRest: bool = (options & self.OPT_NOREST) == self.OPT_NOREST
        optionNoTie: bool = (options & self.OPT_NOTIE) == self.OPT_NOTIE

        for line in self._lines:
            if line.isEmpty:
                continue

            tempTokens = []
            if not optionNoGlobal and line.isGlobal:
                # append [the entire global line] to output
                token0: HumdrumToken | None = line[0]
                if token0 is not None:
                    tempTokens.append(token0)
                    output.append(tempTokens)
                continue

            if optionNoEmpty:
                # if all tokens on line (in this track) are null, continue
                allNull: bool = True
                for token in line.tokens():
                    if token.track != track:
                        continue
                    if not token.isNull:
                        allNull = False
                        break
                if allNull:
                    continue

            foundTrack: bool = False
            for token in line.tokens():
                if token.track != track:
                    continue
                if optionPrimary and foundTrack:
                    continue

                foundTrack = True
                if optionNoInterp and token.isInterpretation:
                    continue
                if optionNoManip and token.isManipulator:
                    continue
                if optionNoNull and token.isNull:
                    continue
                if optionNoComment and token.isComment:
                    continue
                if optionNoRest and token.isRest:
                    continue
                if optionNoTie and token.isSecondaryTiedNote:
                    continue

                tempTokens.append(token)

            if tempTokens:
                output.append(tempTokens)

        return output

    '''
    //////////////////////////////
    //
    // HumdrumFileBase::getTrackStart -- Return the starting exclusive
    //     interpretation for the given track.  Returns NULL if the track
    //     number is out of range.
    '''
    def trackStart(self, track: int | None) -> HumdrumToken | None:
        if track is None:
            return None

        if track < 0:
            track += len(self._trackStarts)

        if track < 1 or track >= len(self._trackStarts):
            return None

        return self._trackStarts[track]

    '''
    //////////////////////////////
    //
    // HumdrumFileBase::getTrackEndCount -- Return the number of ending tokens
    //    for the given track.  Spines must start as a single exclusive
    //    interpretation token.  However, since spines may split and merge,
    //    it is possible that there are more than one termination points for a
    //    track.  This function returns the number of terminations which are
    //    present in a file for any given spine/track.
        Returns 0 if track number is out of range.
    '''
    def trackEndCount(self, track: int) -> int:
        if track < 0:
            track += len(self._trackEnds)
        if track < 1 or track >= len(self._trackEnds):
            return 0

        return len(self._trackEnds[track])

    '''
    //////////////////////////////
    //
    // HumdrumFileBase::getTrackEnd -- Returns a pointer to the terminal manipulator
    //    token for the given track and subtrack.  Sub-tracks are indexed from 0 up
    //    to but not including getTrackEndCount.
    '''
    def trackEnd(self, track: int, subTrack: int) -> HumdrumToken | None:
        if track < 0:
            track += len(self._trackEnds)

        if track < 0:
            return None

        if track > self.maxTrack:
            return None

        if subTrack < 0:
            subTrack += self.trackEndCount(track)

        if subTrack < 0 or subTrack >= len(self._trackEnds[track]):
            return None

        return self._trackEnds[track][subTrack]

    '''
    //////////////////////////////
    //
    // HumdrumFileBase::analyzeNonNullDataTokens -- For null data tokens, indicate
    //    the previous non-null token which the null token refers to.  After
    //    a spine merger, there may be multiple previous tokens, so you would
    //        have to decide on the actual source token on based on subtrack or
    //    sub-spine information.  The function also gives links to the previous/next
    //    non-null tokens, skipping over intervening null data tokens.

        Not called internally.  Called by HumdrumFileStructure.py
    '''
    def analyzeNonNullDataTokens(self) -> bool:
        ptokens: list[HumdrumToken] = []

        # analyze forward tokens:
        for i in range(1, self.maxTrack + 1):
            trackStart: HumdrumToken | None = self._trackStarts[i]
            if trackStart is None:
                # I don't think this can happen, but I'm not sure.
                continue
            if not self.processNonNullDataTokensForTrackForward(trackStart, ptokens):
                return False

        ptokens = []  # starting another recursion

        # analyze backward tokens:
        for i in range(1, self.maxTrack + 1):
            for j in range(0, self.trackEndCount(i)):
                if not self.processNonNullDataTokensForTrackBackward(
                        self._trackEnds[i][j], ptokens):
                    return False

        '''
        // Eventually set the forward and backward non-null data token for
        // tokens in spines for all types of line types.  For now specify
        // the next non-null data token for the exclusive interpretation token.
        // Also this implementation does not consider that the first
        // non-null data tokens may be from multiple split tokens (fix later).

        // This algorithm is probably not right, but good enough for now.
        // There may be missing portions of the file for the analysis,
        // and/or the algorithm is probably retracking tokens in the case
        // of spine splits.
        '''
        stops: list[HumdrumToken] = self.spineStopList
        nexts: HumdrumToken | None = None

        for stop in stops:
            if stop is None:
                continue

            # start at the track stop token and work backwards
            token: HumdrumToken | None = stop
            if t.TYPE_CHECKING:
                # we know token is not None because stop/stops are not Optional
                assert isinstance(token, HumdrumToken)
            if token.isNonNullData:
                nexts = token

            token = token.previousToken0
            while token is not None:
                if nexts is not None:
                    token.addNextNonNullDataToken(nexts)
                if token.isNonNullData:
                    nexts = token
                token = token.previousToken0

        return True

    '''
    //////////////////////////////
    //
    // HumdrumFile::processNonNullDataTokensForTrackBackward -- Helper function
    //    for analyzeNonNullDataTokens.  Given any token, this function tells
    //    you what is the next non-null data token(s) in the spine after the given
    //    token.
    '''
    def processNonNullDataTokensForTrackBackward(self,
        endToken: HumdrumToken,
        ptokens: list[HumdrumToken]
    ) -> bool:
        token: HumdrumToken = endToken
        tcount: int = token.previousTokenCount

        while tcount > 0:
            for i in range(1, tcount):
                prevtok: HumdrumToken | None = token.previousToken(i)
                if t.TYPE_CHECKING:
                    # we know prevtok is not None because i is in range
                    assert isinstance(prevtok, HumdrumToken)

                if not self.processNonNullDataTokensForTrackBackward(prevtok, ptokens):
                    return False

            prevToken: HumdrumToken | None = token.previousToken0
            if t.TYPE_CHECKING:
                # we know prevToken is not None because 0 is in range (tcount > 0)
                assert isinstance(prevToken, HumdrumToken)

            if prevToken.isSplitInterpretation:
                self.addUniqueTokens(prevToken.nextNonNullDataTokens, ptokens)
                if token != prevToken.nextToken0:
                    # terminate if not most primary subspine
                    return True
            elif token.isData:
                self.addUniqueTokens(token.nextNonNullDataTokens, ptokens)
                if not token.isNull:
                    ptokens = [token]  # nukes the existing ptokens list

            # Follow previous data token 0 since 1 and higher are handled above.
            token = prevToken
            tcount = token.previousTokenCount

        return True

    '''
    //////////////////////////////
    //
    // HumdrumFile::processNonNullDataTokensForTrackForward -- Helper function
    //    for analyzeNonNullDataTokens.  Given any token, this function tells
    //    you what are the previous non-null data token(s) in the spine before
    //    the given token.
    '''
    def processNonNullDataTokensForTrackForward(self,
        startToken: HumdrumToken,
        ptokens: list[HumdrumToken]
    ) -> bool:
        token: HumdrumToken = startToken
        tcount: int = token.nextTokenCount

        while tcount > 0:
            nextToken: HumdrumToken | None = None
            if token.isSplitInterpretation:
                for i in range(1, tcount):
                    nextTok: HumdrumToken | None = token.nextToken(i)
                    if t.TYPE_CHECKING:
                        # we know nextTok is not None because i is in range
                        assert isinstance(nextTok, HumdrumToken)

                    if not self.processNonNullDataTokensForTrackForward(nextTok, ptokens):
                        return False
            elif token.isMergeInterpretation:
                nextToken = token.nextToken0
                if t.TYPE_CHECKING:
                    # we know nextToken is not None because 0 is in range (tcount > 0)
                    assert isinstance(nextToken, HumdrumToken)

                self.addUniqueTokens(nextToken.previousNonNullDataTokens, ptokens)
                if token != nextToken.previousToken0:
                    # terminate if not most primary subspine
                    return True
            else:
                self.addUniqueTokens(token.previousNonNullDataTokens, ptokens)
                if token.isNonNullData:
                    ptokens = [token]  # nukes the existing ptokens list

            # Data tokens can only be followed by up to one next token,
            # so no need to check for more than one next token.
            nextToken = token.nextToken0
            if t.TYPE_CHECKING:
                # we know nextToken is not None because 0 is in range (tcount > 0)
                assert isinstance(nextToken, HumdrumToken)

            token = nextToken
            tcount = token.nextTokenCount

        return True

    '''
    //////////////////////////////
    //
    // operator<< -- Default method of printing HumdrumFiles.  This printing method
    //    assumes that the HumdrumLine string is correct.  If a token is changed
    //    in the file, the HumdrumFileBase::createLinesFromTokens() before printing
    //    the contents of the line.
    '''
    def write(self, fp) -> None:
        for line in self._lines:
            line.write(fp)



import pytest

# The things we're testing
from converter21.humdrum import HumdrumLine

# test utilities
from tests.Utilities import *

def test_HumdrumLine_default_init():
    line = HumdrumLine()
    line.createTokensFromLine()
    CheckHumdrumLine(line)

def test_HumdrumLine_single_exinterp():
    line = HumdrumLine('**kern')
    line.createTokensFromLine()
    CheckHumdrumLine(line, expectedLine = '**kern',
                           expectedLineNumber = 0,
                           expectedType = LINETYPE_INTERPRETATION,
                           expectedTokenCount = 1,
                           expectedIsExclusiveInterpretation = True,
                           expectedIsManipulator = True,
                           expectedTokens = ['**kern'])

def test_HumdrumLine_manipulators():
    line = HumdrumLine('*\t*^\t*-\t*')
    line.createTokensFromLine()
    CheckHumdrumLine(line, expectedLine = '*\t*^\t*-\t*',
                           expectedLineNumber = 0,
                           expectedType = LINETYPE_INTERPRETATION,
                           expectedTokenCount = 4,
                           expectedIsExclusiveInterpretation = False,
                           expectedIsManipulator = True,
                           expectedTokens = ['*', '*^', '*-', '*'])

def test_HumdrumLine_barline():
    line = HumdrumLine('=1\t=1\t=1')
    line.createTokensFromLine()
    CheckHumdrumLine(line, expectedLine = '=1\t=1\t=1',
                           expectedLineNumber = 0,
                           expectedType = LINETYPE_BARLINE,
                           expectedTokenCount = 3,
                           expectedIsExclusiveInterpretation = False,
                           expectedIsManipulator = False,
                           expectedTokens = ['=1', '=1', '=1'])

def test_HumdrumLine_data():
    line = HumdrumLine('4f]\t.\t2D\t(1cc\t.')
    line.createTokensFromLine()
    CheckHumdrumLine(line, expectedLine = '4f]\t.\t2D\t(1cc\t.',
                           expectedLineNumber = 0,
                           expectedType = LINETYPE_DATA,
                           expectedTokenCount = 5,
                           expectedIsExclusiveInterpretation = False,
                           expectedIsManipulator = False,
                           expectedTokens = ['4f]', '.', '2D', '(1cc', '.'])

# add more tests for coverage...
